/*
 This file is part of iDQ.

 iDQ is an instantiation-based DQBF (Dependency Quantified Boolean Formula) solver.

 Andreas Froehlich, Johannes Kepler University, Linz, Austria, 2012-2014.
 Gergely Kovasznai, Vienna University of Technology, Vienna, Austria, 2012-2014.

 iDQ is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or (at
 your option) any later version.

 iDQ is distributed in the hope that it will be useful, but
 WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with iDQ.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdio.h>

#include "abformula.h"
#include "qbformula.h"
#include "bformula.h"
#include "satsolver.h"
#include "mem.h"
#include "assert.h"

/*
 * Comparison functions for the priority queue of passive instances
 */

// for default heuristics
static int cmp_default(const void * p, const void * q) {
	PassiveInstRef * a = (PassiveInstRef *) p;
	PassiveInstRef * b = (PassiveInstRef *) q;

	int res;

	// +ground
	/*	if ((res = !a->numDeps - !b->numDeps))
	 return res;*/

	// -num_deps
	if ((res = a->numDeps - b->numDeps))
		return -res;

	/*	// -num_lits
	 if ((res = a->numLits - b->numLits))
	 return -res;*/

	// +age
	if ((res = b->createdAt - a->createdAt))
		return res;

	// -num_symbs
	((res = a->numSymbs - b->numSymbs));
	return -res;
}

// for VSIDS heuristics
static int cmp_byVarScore(const void * p, const void * q) {
	PassiveInstRef * a = (PassiveInstRef *) p;
	PassiveInstRef * b = (PassiveInstRef *) q;

	int res;

//	if (rand() % 2 == 0) {
	// average variable score
	AbClauseInst *a_inst = GET_INST_BY_INSTREF(a->instRef), *b_inst = GET_INST_BY_INSTREF(b->instRef);
	float a_score = a_inst->varScoreSum / (float) a_inst->bClause->cntLits, b_score = b_inst->varScoreSum / (float) b_inst->bClause->cntLits;
	if ((res = (a_score > b_score) - (a_score < b_score)))
		return res;
//	}

	return cmp_default(a, b);
}

/*
 * Initializes the abstract vars:
 * 		- Sets its dependencies
 * 		- Sets the literal watchers
 */
static void init_vars(iDQMemMan * mm, AbFormula * abf, QBFormula * qbf) {
	AbVariable * abVar;
	Variable * qbfVar;

	unsigned int * pDep;
	unsigned int varId, i;
	for (varId = 1; varId <= abf->cntVars; varId++) {
		abVar = abf->vars + varId;
		qbfVar = qbf->pVars + abf->pExMap[varId];

		abVar->depVarMask = create_iList(mm, qbf->cntUniversals, UNSIGNED);

		for (pDep = qbfVar->dependencies, i = qbfVar->cntDependencies; i; pDep++, i--)
			setLit_iList(abVar->depVarMask, abf->pVarMap[*pDep]);

		INIT_STACK(mm, abVar->posLitWatchers);
		INIT_STACK(mm, abVar->negLitWatchers);
	}
}

/*
 * Basic initialization of an AbFormula
 */
static void basicInit(iDQ * idq) {
	iDQMemMan * mm = idq->mm;
	AbFormula * abf = idq->abf;
	QBFormula * qbf = idq->qbf;

	abf->cntVars = qbf->cntExistentials;
	abf->cntClauses = qbf->cntClauses - qbf->cntDeletedClauses;
	abf->cntInstances = 0;

	abf->cntUniversals = qbf->cntUniversals;
	abf->cntExistentials = qbf->cntExistentials;

	NEWN(mm, abf->vars, abf->cntVars + 1);
	NEWN(mm, abf->clauses, abf->cntClauses);

	NEWN(mm, abf->pVarMap, qbf->cntVars + 1);
	NEWN(mm, abf->pExMap, qbf->cntExistentials + 1);
	NEWN(mm, abf->pUnivMap, qbf->cntUniversals + 1);

	int exMapPos = 1;
	int univMapPos = 1;

	unsigned int varId;
	Variable * currVar;

	for (varId = 1; varId <= qbf->cntVars; varId++) {
		currVar = qbf->pVars + varId;

		if (currVar->quantifier == QTYPE_EXISTS) {
			//printf("Writing %d in exMap at pos %d\n", i, exMapPos);
			abf->pExMap[exMapPos] = varId;
			abf->pVarMap[varId] = exMapPos++;
		} else if (currVar->quantifier == QTYPE_FORALL) {
			//printf("Writing %d in univMap at pos %d\n", i, univMapPos);
			abf->pUnivMap[univMapPos] = varId;
			abf->pVarMap[varId] = univMapPos++;
		}
	}

	init_vars(mm, abf, qbf);

	abf->passiveInsts = create_priorQueue(mm, abf, (idq->inst_sel_mode ? cmp_byVarScore : cmp_default));
}

/*
 * Deletes an AbClause
 */
static void delete_abClauseInst(iDQ * idq, AbClause * abClause, AbClauseInst * inst) {
	delete_iList(inst->instMask);
	delete_iList(inst->groundInstMask);
	if (inst->bClause)
		delete_bClause(idq->mm, inst->bClause);
#ifdef EXTEND_REDUNDANCY_CHECK_WITH_TIME
	DELETEN(idq->mm, inst->pSelLitFrom, abClause->cntLits);
	DELETEN(idq->mm, inst->pSelLitUntil, abClause->cntLits);
#endif
	DELETE(idq->mm, inst->bClause);

#ifdef INST_VARSCORE_ON_THE_FLY
	if (USE_VARSCORE_HEURISTICS(idq) && inst->offsets_instLitWatchers)
		DELETEN(idq->mm, inst->offsets_instLitWatchers, abClause->cntLits);
#endif
}

/*
 * Deletes an AbClause
 */
static void delete_abClause(iDQ * idq, AbClause * abClause) {
	DELETEN(idq->mm, abClause->lits, abClause->cntLits);

	unsigned int i;
	AbClauseInst * pInst;
	for (pInst = abClause->instances.start, i = COUNT_STACK(abClause->instances); i; pInst++, i--)
		delete_abClauseInst(idq, abClause, pInst);
	DELETE_STACK(abClause->instances);

	delete_iList(abClause->depVarMask);
}

/*
 * Deletes an AbFormula
 */
void delete_abFormula(iDQ * idq, AbFormula * abf) {
	unsigned int i;

// delete all the clauses
	AbClause * pAbClause;
	for (i = abf->cntClauses, pAbClause = abf->clauses; i; i--, pAbClause++)
		delete_abClause(idq, pAbClause);
	DELETEN(idq->mm, abf->clauses, abf->cntClauses);

// delete all the variables
	AbVariable * pAbVar;
	for (i = abf->cntVars + 1, pAbVar = abf->vars; i; i--, pAbVar++) {
		if (pAbVar->depVarMask)
			delete_iList(pAbVar->depVarMask);
		DELETE_STACK(pAbVar->posLitWatchers);
		DELETE_STACK(pAbVar->negLitWatchers);
	}
	DELETEN(idq->mm, abf->vars, abf->cntVars + 1);

// delete the variable maps
	DELETEN(idq->mm, abf->pVarMap, abf->cntVars + 1);
	DELETEN(idq->mm, abf->pUnivMap, abf->cntUniversals + 1);
	DELETEN(idq->mm, abf->pExMap, abf->cntExistentials + 1);

// delete the collection of passive instances
	delete_priorQueue(abf->passiveInsts);
}

/*
 * Checks if a (candidate) instance already exists among the instances
 */
static int exists_inst(AbFormula * abf, AbClause * abClause, IList* newInstMask) {
	ASSERT(newInstMask);

	IList * groundNewInstMask = union_iList(newInstMask, abClause->depVarMask);

	unsigned int i;
	char exists = 0;
	AbClauseInst *pInst;
	unsigned int cntInsts = COUNT_STACK(abClause->instances), cntActiveInsts = abClause->cntActiveInsts;
	for (pInst = abClause->instances.start, i = 0; i < cntInsts; pInst++, i++) {
		if (equal_iList(pInst->groundInstMask, groundNewInstMask, SIGNED)) {
			if (subseteq_iList(pInst->instMask, newInstMask, SIGNED)) {
				delete_iList(newInstMask);
				exists = 1;
				break;
			}
#ifdef PREFER_GENERAL_INSTANCES
			else if (subseteq_iList(newInstMask, pInst->instMask, SIGNED)) {
				IList * oldInstMask = pInst->instMask;
				pInst->instMask = newInstMask;
				delete_iList(oldInstMask);

				// if a passive inst has been replaced, update its heuristic data and rearrange it in the priority queue of passive insts
				if (i >= cntActiveInsts) {
					ASSERT(abf->passiveInsts->list[pInst->offset_instWatcher]->instRef.abClause == abClause && abf->passiveInsts->list[pInst->offset_instWatcher]->instRef.instInd == i);
					COMPUTE_PASSIVE_INST_HEUR_DATA(abf->passiveInsts->list[pInst->offset_instWatcher]);
					rearrangeElement_priorQueue(abf->passiveInsts, pInst->offset_instWatcher);
				}

				exists = 1;
				break;
			}
#endif
		}
	}

	delete_iList(groundNewInstMask);
	return exists;
}
/*
 * possible improvement: first (before iterating through the instances) check if the ground new instance equals to the ground old instance
 */

/*
 * Adds a new instances to an AbClause
 */
static void add_inst(iDQ * idq, AbClause * abClause, IList* newInstMask) {
	AbClauseInst newInst;

	newInst.instMask = newInstMask;
	newInst.groundInstMask = union_iList(newInstMask, abClause->depVarMask);

	newInst.selLitInd = INVALID_LIT_IND;
#ifdef EXTEND_REDUNDANCY_CHECK_WITH_TIME
	CNEWN(idq->mm, newInst.pSelLitFrom, abClause->cntLits);
	CNEWN(idq->mm, newInst.pSelLitUntil, abClause->cntLits);
#endif

	newInst.createdAt = idq->clock;

	PUSH_STACK(abClause->instances, newInst);
	idq->abf->cntInstances++;

	unsigned int pushed_instInd = COUNT_STACK(abClause->instances) - 1;
	AbClauseInst * pushed_inst = GET_INST_BY_ABCLAUSE_AND_INSTIND(abClause, pushed_instInd);

	if (idq->inst_sel_mode) {
		pushed_inst->varScoreSum = 0;
#ifdef INST_VARSCORE_ON_THE_FLY
		if (USE_VARSCORE_HEURISTICS(idq)) {
			if (IS_INST_PASSIVE(abClause, pushed_instInd))
				NEWN(idq->mm, pushed_inst->offsets_instLitWatchers, abClause->cntLits);
			else
				pushed_inst->offsets_instLitWatchers = NULL;
		}
#endif
	}

	BClause * bClause = convert_instToBClause(idq, abClause, COUNT_STACK(abClause->instances) - 1);
	pushed_inst->bClause = bClause;

	if (idq->inst_sel_mode && IS_INST_PASSIVE(abClause, pushed_instInd)) {
		unsigned int i;
		for (i = 0; i < abClause->cntLits; i++)
			pushed_inst->varScoreSum += bClause->pVars[i]->score;
	}
}

/*
 * Pushes a new passive instance into the priority queue
 */
static void push_passiveInst(iDQ * idq, AbClause * abClause, IList * instMask) {
	AbFormula * abf = idq->abf;

	if (exists_inst(abf, abClause, instMask))
		return;

	add_inst(idq, abClause, instMask);

	PassiveInstRef * newPassiveInst;
	NEW(idq->mm, newPassiveInst);
	newPassiveInst->instRef.abClause = abClause;
	newPassiveInst->instRef.instInd = COUNT_STACK(abClause->instances) - 1;
	newPassiveInst->createdAt = idq->clock;
	COMPUTE_PASSIVE_INST_HEUR_DATA(newPassiveInst);

// push the new passive inst, and stores its offset in the priority queue of passive insts
	GET_INST_BY_INSTREF(newPassiveInst->instRef)->offset_instWatcher = enqueue_priorQueue(abf->passiveInsts, newPassiveInst);
}

#ifdef INST_VARSCORE_ON_THE_FLY
/*
 * Removes an inst from the var watcher lists
 */
static void remove_varWatchers(AbFormula * abf, AbClause * abClause, unsigned int instInd) {
	AbClauseInst * inst = GET_INST_BY_ABCLAUSE_AND_INSTIND(abClause, instInd);

	InstLitRefStack * watcherList;
	InstLitRef popped_instLitRef;
	AbClauseInst * popped_inst;
	unsigned int i;
	for (i = 0; i < abClause->cntLits; i++) {
		watcherList = &(inst->bClause->pVars[i]->varWatchers);
		ASSERT(!EMPTY_STACK(*watcherList));

		// pop (the last) entry from the watcher list...
		popped_instLitRef = POP_STACK(*watcherList);
		popped_inst = GET_INST_BY_ABCLAUSE_AND_INSTIND(popped_instLitRef.abClause, popped_instLitRef.instInd);

		// rewrite the list entry being removed with the popped entry...
		*GET_STACK(*watcherList, inst->offsets_instLitWatchers[i]) = popped_instLitRef;

		//...and update the appropriate watcher list offset of the popped entry
		popped_inst->offsets_instLitWatchers[popped_instLitRef.litInd] = inst->offsets_instLitWatchers[i];
	}
}
#endif

#ifdef ALLOW_TO_MOVE_ACTIVE_TO_PASSIVE
static void update_litWacher_instInd(AbFormula * abf, AbClause * abClause, unsigned int instInd, unsigned int newInstInd) {
	AbClauseInst * inst = GET_INST_BY_ABCLAUSE_AND_INSTIND(abClause, instInd);

	if (inst->selLitInd == INVALID_LIT_IND)
	return;

	LitID selLitId = abClause->lits[inst->selLitInd];
	AbVariable * var = abf->vars + GET_VARID(selLitId);

	// w.r.t. the sign of the watched literal, the inst must be removed from the watcher list either posLiteralWatchers or negLiteralWatchers
	InstRefStack * watcherList = (selLitId > 0 ? &var->posLitWatchers : &var->negLitWatchers);

	// update the watchers of the last active inst
	GET_STACK(*watcherList, inst->offset_instWatcher)->instInd = newInstInd;
}
#endif

#define SWAP_INSTS(x, y) 	\
{	\
	AbClauseInst temp = *(x);		\
	*(x) = *(y);		\
	*(y) = temp;		\
}

/*
 * Pops appropriate number of passive instances from the priority queue
 */
unsigned int pop_passiveInsts(iDQ * idq) {
	iDQMemMan * mm = idq->mm;
	AbFormula * abf = idq->abf;

	InstRefStack popped_satInsts, popped_otherInsts;
	INIT_STACK(mm, popped_satInsts);
	INIT_STACK(mm, popped_otherInsts);

	PassiveInstRef * passiveInstRef;
	AbClause * abClause;
	AbClauseInst *first_passiveInst, *popped_inst, temp;
	InstRef popped_instRef;

	unsigned int i;
	BVariable * bVar;
#ifdef INST_VARSCORE_ON_THE_FLY
	InstLitRef * pInstLitRef;
#endif

#ifdef USE_BATCH_PROCESSING_NEW_INSTANCES
// pop several instances
	while (COUNT_STACK(popped_otherInsts) < MAX_BATCH_SIZE && COUNT_STACK(popped_otherInsts) * BATCH_SIZE_RATE < abf->passiveInsts->count) {
//	while (abf->passiveInsts->count && cntNewActiveInsts < MAX_BATCH_SIZE) {
#else
		// pop only 1 instance
		while (abf->passiveInsts->count && EMPTY_STACK(popped_otherInsts)) {
#endif
		passiveInstRef = dequeue_priorQueue(abf->passiveInsts);

		abClause = passiveInstRef->instRef.abClause;
		popped_inst = GET_INST_BY_INSTREF(passiveInstRef->instRef);

#ifdef INST_VARSCORE_ON_THE_FLY
		if (USE_VARSCORE_HEURISTICS(idq)) {
			// remove the popped inst from the var watcher lists
			remove_varWatchers(abf, abClause, passiveInstRef->instRef.instInd);
			DELETEN(mm, popped_inst->offsets_instLitWatchers, abClause->cntLits);
			popped_inst->offsets_instLitWatchers = NULL;
		}
#endif

// swap the popped inst with the first passive inst
		first_passiveInst = GET_INST_BY_ABCLAUSE_AND_INSTIND(abClause, abClause->cntActiveInsts);
		if (first_passiveInst != popped_inst) {
			// update the watchers of the first passive inst
			ASSERT(abf->passiveInsts->list[first_passiveInst->offset_instWatcher]->instRef.instInd == abClause->cntActiveInsts);
			abf->passiveInsts->list[first_passiveInst->offset_instWatcher]->instRef.instInd = passiveInstRef->instRef.instInd;

#ifdef INST_VARSCORE_ON_THE_FLY
			if (USE_VARSCORE_HEURISTICS(idq)) {
				for (i = 0; i < abClause->cntLits; i++) {
					bVar = first_passiveInst->bClause->pVars[i];
					GET_STACK(bVar->varWatchers, first_passiveInst->offsets_instLitWatchers[i])->instInd = passiveInstRef->instRef.instInd;
				}
			}
#endif
			SWAP_INSTS(first_passiveInst, popped_inst);
		}

		popped_instRef.abClause = abClause;
		popped_instRef.instInd = abClause->cntActiveInsts++;
		if (getValue_inst(popped_instRef.abClause, popped_instRef.instInd, idq->satSolver) == RESULT_SAT)
			PUSH_STACK(popped_satInsts, popped_instRef)
		else
			PUSH_STACK(popped_otherInsts, popped_instRef)

		DELETE(mm, passiveInstRef);
	}

	unsigned int res = COUNT_STACK(popped_satInsts) + COUNT_STACK(popped_otherInsts);
	if (idq->verbosity_level > 1)
		printf("#popped instances = %d (#sat = %d)\n", res, COUNT_STACK(popped_satInsts));

	while (!EMPTY_STACK(popped_satInsts)) {
		popped_instRef = POP_STACK(popped_satInsts);
		add_inst_toSatSolver(idq, popped_instRef.abClause, popped_instRef.instInd);
	}
	DELETE_STACK(popped_satInsts);

	while (!EMPTY_STACK(popped_otherInsts)) {
		popped_instRef = POP_STACK(popped_otherInsts);
		add_inst_toSatSolver(idq, popped_instRef.abClause, popped_instRef.instInd);
	}
	DELETE_STACK(popped_otherInsts);

	return res;
}

/*
 * Initializes an AbClause, by "copying" literals to it from a temporary stack
 */
static void init_abClause(iDQ * idq, AbClause * abClause) {
	AbFormula * abf = idq->abf;

	abClause->cntLits = COUNT_STACK(idq->tempLitStack);
	NEWN(idq->mm, abClause->lits, abClause->cntLits);

	LitID *pTempLit, *pEndTemp = idq->tempLitStack.top, *pLit;
	for (pTempLit = idq->tempLitStack.start, pLit = abClause->lits; pTempLit != pEndTemp; pTempLit++, pLit++)
		*pLit = (*pTempLit > 0 ? 1 : -1) * abf->pVarMap[GET_VARID(*pTempLit)];

	DELETE_STACK(idq->tempLitStack);

	INIT_STACK(idq->mm, abClause->instances);
}

/*
 * Adds the 1st instance to an AbClause, by assigning all its universal literals to false
 * (the universal literals have been collected in a temporary stack before)
 */
static void add_firstInst(iDQ * idq, AbClause * abClause) {
	AbFormula * abf = idq->abf;

	IList * instMask = create_iList(idq->mm, abf->cntUniversals, SIGNED);

	unsigned int varId;
	LitID sign;
	LitID *pTempLit;
	unsigned int i;
	for (pTempLit = idq->tempLitStack.start; pTempLit != idq->tempLitStack.top; pTempLit++) {
		varId = abf->pVarMap[GET_VARID(*pTempLit)];
		sign = (*pTempLit > 0 ? -1 : 1);

		// check if the (universal) var has already added with opposite sign => tautological clause, no need to add
		if (getLit_iList(instMask, varId) == -sign) {
			delete_iList(instMask);
			DELETE_STACK(idq->tempLitStack);
			return;
		}

		setLit_iList(instMask, sign * varId);
	}

	DELETE_STACK(idq->tempLitStack);

	intersectWith_iList(instMask, abClause->depVarMask);

// add an active instance
	abClause->cntActiveInsts = 1;
	add_inst(idq, abClause, instMask);
	add_inst_toSatSolver(idq, abClause, 0);
}

/*
 * "Abstracts" the original clause:
 * 		- Gives a new 1-based index to existential variables
 * 		- Finds out the dependencies of each clause
 * 		- Generates the 1st instance of each clause
 */
void create_abFormula(iDQ * idq) {
	iDQMemMan * mm = idq->mm;
	QBFormula * qbf = idq->qbf;

	NEW(mm, idq->abf);
	AbFormula * abf = idq->abf;

	basicInit(idq);

	unsigned int clauseIndex;
	Variable * currVar;
	IList * depVarMask;
	AbClause * pAbClause;
	Clause * pClause;
	LitID * pLit;
	unsigned int * pDep;
	unsigned int i, j;
	for (pClause = qbf->pClauses, pAbClause = abf->clauses, clauseIndex = 0; clauseIndex < qbf->cntClauses; pClause++, clauseIndex++) {
		if (pClause->is_deleted)
			continue;

		depVarMask = create_iList(mm, abf->cntUniversals, UNSIGNED);

		// collect all the existential vars and collect all their dependencies
		for (pLit = pClause->lits, i = pClause->cntLits; i; pLit++, i--) {
			currVar = qbf->pVars + GET_VARID(*pLit);

			if (currVar->quantifier == QTYPE_EXISTS) {
				for (pDep = currVar->dependencies, j = currVar->cntDependencies; j; pDep++, j--)
					setLit_iList(depVarMask, abf->pVarMap[*pDep]);

				PUSH_STACK(idq->tempLitStack, *pLit);
			}
		}

		// create a new abClause that correspond to the original clause, and set its dependencies
		init_abClause(idq, pAbClause);
		pAbClause->depVarMask = depVarMask;

		// collect all the universal vars
		for (pLit = pClause->lits, i = pClause->cntLits; i; pLit++, i--) {
			currVar = qbf->pVars + GET_VARID(*pLit);

			if (currVar->quantifier == QTYPE_FORALL)
				PUSH_STACK(idq->tempLitStack, *pLit);
		}

		// generate the 1st instance of the AbClause
		add_firstInst(idq, pAbClause);

		pAbClause++;
	}
}

/*
 * Resolves two instances
 */
static void resolve_twoInsts(iDQ * idq, AbClause * abClause1, AbClauseInst * inst1, AbClause * abClause2, AbClauseInst * inst2, unsigned int varId) {
	iDQMemMan * mm = idq->mm;
	AbFormula * abf = idq->abf;

	AbVariable * abstVar = abf->vars + varId;
	IList * abstVarMask = abstVar->depVarMask;

	IList * instMask1 = inst1->instMask;
	IList * instMask2 = inst2->instMask;

	IList * lit1Mask = intersection_iList(instMask1, abstVarMask);
	IList * lit2Mask = intersection_iList(instMask2, abstVarMask);

	if (equal_iList(lit1Mask, lit2Mask, UNSIGNED)) {
		// cannot instantiate clauses because the are equally strong
		delete_iList(lit1Mask);
		delete_iList(lit2Mask);
		return;
	}

	IList * intersectionLitMask = intersection_iList(lit1Mask, lit2Mask);
	if (!intersectionLitMask) {
		// cannot instantiate clauses because they contain dual variables
		delete_iList(lit1Mask);
		delete_iList(lit2Mask);
		return;
	}
	delete_iList(intersectionLitMask);

	IList * unionInstMask;
	if (!subseteq_iList(lit2Mask, instMask1, SIGNED) && (unionInstMask = union_iList(lit2Mask, instMask1)))
		push_passiveInst(idq, abClause1, unionInstMask);
	if (!subseteq_iList(lit1Mask, instMask2, SIGNED) && (unionInstMask = union_iList(lit1Mask, instMask2)))
		push_passiveInst(idq, abClause2, unionInstMask);

	delete_iList(lit1Mask);
	delete_iList(lit2Mask);
}

/*
 * Update the literal watchers from an old literal index to a new one
 */
static void update_litWatcher(AbFormula * abf, AbClause * abClause, unsigned int instInd, unsigned int newSelLitInd) {
//	ASSERT(instInd < abClause->cntActiveInsts);
//	ASSERT(newSelLitInd < abClause->cntLits);

	AbClauseInst * inst = GET_INST_BY_ABCLAUSE_AND_INSTIND(abClause, instInd);

	unsigned int selLitInd = inst->selLitInd;
	LitID selLitId;
	AbVariable * var;
	InstRefStack * watcherList;

	if (selLitInd != INVALID_LIT_IND) { // remove inst from the watcher list of the old selected lit
		selLitId = abClause->lits[selLitInd];
		var = abf->vars + GET_VARID(selLitId);

		// w.r.t. the sign of the watched literal, the inst must be removed from the watcher list either posLiteralWatchers or negLiteralWatchers
		watcherList = (selLitId > 0 ? &var->posLitWatchers : &var->negLitWatchers);
		ASSERT(!EMPTY_STACK(*watcherList));

		// pop (the last) inst from the watcher list (i.e., stack)...
		InstRef poppedInstRef = POP_STACK(*watcherList);
		AbClauseInst * poppedInst = GET_INST_BY_INSTREF(poppedInstRef);
		ASSERT(poppedInst->selLitInd != INVALID_LIT_IND);
		ASSERT(poppedInstRef.abClause->lits[poppedInst->selLitInd] == selLitId);

		// rewrite the (list entry of the) inst being removed with the popped inst...
		ASSERT(GET_STACK(*watcherList, inst->offset_instWatcher)->abClause == abClause && GET_STACK(*watcherList, inst->offset_instWatcher)->instInd == instInd);
		*GET_STACK(*watcherList, inst->offset_instWatcher) = poppedInstRef;

		//...and update the appropriate watcher list offset of the popped inst
		poppedInst->offset_instWatcher = inst->offset_instWatcher;
	}

	if (newSelLitInd == INVALID_LIT_IND)
		return;

	selLitId = abClause->lits[newSelLitInd];
	var = abf->vars + GET_VARID(selLitId);

// w.r.t. the sign of the watched literal, the clause must be added to the watcher list either posLiteralWatchers or negLiteralWatchers
	watcherList = (selLitId > 0 ? &var->posLitWatchers : &var->negLitWatchers);

// push the clause to the appropriate watcher list
	inst->offset_instWatcher = COUNT_STACK(*watcherList);
	InstRef newInstRef;
	newInstRef.abClause = abClause;
	newInstRef.instInd = instInd;
	PUSH_STACK(*watcherList, newInstRef);
}

#ifdef EXTEND_REDUNDANCY_CHECK_WITH_TIME
// to check if a value falls into the interval [left,right)
#define FALLS_INTO_INTERVAL(val, left, right) ((val) >= (left) && (val) < (right))
#endif

/*
 * Resolves an instance with all the instances
 */
static void resolve_inst_withAllInsts(iDQ * idq, AbClause * givenAbClause, unsigned int givenInstInd) {
	AbFormula * abf = idq->abf;
	BFormula * bf = idq->bf;

	unsigned int givenLitInd = GET_INST_BY_ABCLAUSE_AND_INSTIND(givenAbClause, givenInstInd)->selLitInd;
	LitID givenLitId = givenAbClause->lits[givenLitInd];

	unsigned int varId = GET_VARID(givenLitId);
	AbVariable * var = abf->vars + varId;
	InstRefStack * watcherList = (givenLitId > 0 ? &var->negLitWatchers : &var->posLitWatchers);
	InstRef * pInstRef;
	AbClauseInst *givenInst, *inst;
	for (pInstRef = watcherList->start; pInstRef != watcherList->top; pInstRef++) {
		givenInst = GET_INST_BY_ABCLAUSE_AND_INSTIND(givenAbClause, givenInstInd);
		inst = GET_INST_BY_INSTREF(*pInstRef);

		if (inst->isGiven)
			continue;

		ASSERT(pInstRef->abClause->lits[inst->selLitInd] == -givenLitId);

		// if the instances have not been resolved on the same literals yet
		// (if the inst's lit selection interval overlaps with the (old) interval of the givenInst)
#ifdef EXTEND_REDUNDANCY_CHECK_WITH_TIME
		if (!givenInst->pSelLitUntil[givenLitInd]
				|| (!FALLS_INTO_INTERVAL(inst->pSelLitFrom[inst->selLitInd], givenInst->pSelLitFrom[givenLitInd], givenInst->pSelLitUntil[givenLitInd])
						&& !FALLS_INTO_INTERVAL(inst->pSelLitUntil[inst->selLitInd], givenInst->pSelLitFrom[givenLitInd] + 1, givenInst->pSelLitUntil[givenLitInd] + 1)))
#endif
		resolve_twoInsts(idq, givenAbClause, givenInst, pInstRef->abClause, inst, varId);
	}
}

// push an instance ref to a stack
#define PUSH_INSTREF(_instRefStack, _abClause, _instInd)	\
{	\
	InstRef instRef;	\
	instRef.abClause = (_abClause);	\
	instRef.instInd = (_instInd);	\
	PUSH_STACK((_instRefStack), instRef);		\
}

/*
 * Generates new passive (i.e., candidate) instances
 */
void generate_passiveInsts(iDQ * idq) {
	AbFormula * abf = idq->abf;
	BFormula * bf = idq->bf;

// update all variable scores (only if VSIDS used)
	if (USE_VARSCORE_HEURISTICS(idq) && idq->var_score_mode)
		UPDATE_VAR_SCORES(idq);

// stack of the "given" (i.e., chosen) insts
	InstRefStack givenInsts;
	INIT_STACK(idq->mm, givenInsts);

	unsigned int litInd;

	AbClause * pAbClause;
	AbClauseInst * pInst;
	unsigned int cntInsts;
	unsigned int i, j, k;

	if (idq->inst_sel_mode) {
#if !defined(INST_VARSCORE_ON_THE_FLY)
// iterate through all the passive instances, and update their varScoreSum
		for (pAbClause = abf->clauses, i = abf->cntClauses; i; pAbClause++, i--) {
			cntInsts = COUNT_STACK(pAbClause->instances);
			for (j = pAbClause->cntActiveInsts; j < cntInsts; j++) {
				pInst = GET_INST_BY_ABCLAUSE_AND_INSTIND(pAbClause, j);

				pInst->varScoreSum = 0;
				for (k = 0; k < pAbClause->cntLits; k++)
				pInst->varScoreSum += pInst->bClause->pVars[k]->score;

				rearrangeElement_priorQueue(abf->passiveInsts, pInst->offset_instWatcher);
			}
		}
#endif
	}

// iterate through all the active instances, and check if literal selection has been changed for the instance
// if it has, then mark the instance as "given"
	for (pAbClause = abf->clauses, i = abf->cntClauses; i; pAbClause++, i--)
		for (j = 0; j < pAbClause->cntActiveInsts; j++) {
			pInst = GET_INST_BY_ABCLAUSE_AND_INSTIND(pAbClause, j);

			if (!idq->lit_sel_mode) {
				// if the selected lit is still true, then do not change the lit selection
				if (pInst->selLitInd != INVALID_LIT_IND && getValue_lit(pAbClause, j, pInst->selLitInd, idq->satSolver) == RESULT_SAT)
					continue;
			}

			litInd = select_satLiteral(idq, pAbClause, j);

			if (pInst->selLitInd != litInd) {
#ifdef EXTEND_REDUNDANCY_CHECK_WITH_TIME
				if (pInst->selLitInd != INVALID_LIT_IND)
				pInst->pSelLitUntil[pInst->selLitInd] = idq->clock; // remember that the selection period of the "old" lit is closing
#endif

#ifdef ALLOW_TO_MOVE_ACTIVE_TO_PASSIVE
				if (pInst->selLitInd == INVALID_LIT_IND) {
#endif

				update_litWatcher(abf, pAbClause, j, litInd);
				pInst->selLitInd = litInd;

				// push the inst to the stack of given insts
				PUSH_INSTREF(givenInsts, pAbClause, j);
				pInst->isGiven = 1;

#ifdef ALLOW_TO_MOVE_ACTIVE_TO_PASSIVE
			} else {
				pAbClause->cntActiveInsts--;

				AbClauseInst * first_passiveInst = GET_INST_BY_ABCLAUSE_AND_INSTIND(pAbClause, pAbClause->cntActiveInsts);
				if (first_passiveInst != pInst) {
					update_litWacher_instInd(abf, pAbClause, pAbClause->cntActiveInsts, j);
					update_litWacher_instInd(abf, pAbClause, j, pAbClause->cntActiveInsts);
					SWAP_INSTS(first_passiveInst, pInst);

					j--;
				}

				// init the selected lit watcher
				update_litWatcher(abf, pAbClause, pAbClause->cntActiveInsts, INVALID_LIT_IND);
				first_passiveInst->selLitInd = INVALID_LIT_IND;

				PassiveInstRef * newPassiveInst;
				NEW(idq->mm, newPassiveInst);
				newPassiveInst->instRef.abClause = pAbClause;
				newPassiveInst->instRef.instInd = pAbClause->cntActiveInsts;
				newPassiveInst->createdAt = idq->clock;
				COMPUTE_PASSIVE_INST_HEUR_DATA(newPassiveInst);

				// push the new passive inst, and stores its offset in the priority queue of passive insts
				GET_INST_BY_INSTREF(newPassiveInst->instRef)->offset_instWatcher = enqueue_priorQueue(abf->passiveInsts, newPassiveInst);
			}
#endif
			}
		}

// iterate through all the given insts, and resolve them with all the active insts
	InstRef * pInstRef;
	for (pInstRef = givenInsts.start; pInstRef != givenInsts.top; pInstRef++) {
		resolve_inst_withAllInsts(idq, pInstRef->abClause, pInstRef->instInd);
		GET_INST_BY_INSTREF(*pInstRef)->isGiven = 0;
	}

#ifdef EXTEND_REDUNDANCY_CHECK_WITH_TIME
	InstRef instRef;
	while (!EMPTY_STACK(givenInsts)) {
		instRef = POP_STACK(givenInsts);
		pInst = GET_INST_BY_INSTREF(instRef);
		pInst->pSelLitFrom[pInst->selLitInd] = idq->clock;
	}
#endif

	DELETE_STACK(givenInsts);
}
