/*
 This file is part of iDQ.

 iDQ is an instantiation-based DQBF (Dependency Quantified Boolean Formula) solver.

 Andreas Froehlich, Johannes Kepler University, Linz, Austria, 2012-2014.
 Gergely Kovasznai, Vienna University of Technology, Vienna, Austria, 2012-2014.

 iDQ is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or (at
 your option) any later version.

 iDQ is distributed in the hope that it will be useful, but
 WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with iDQ.  If not, see <http://www.gnu.org/licenses/>.
*/

/*
 * Data structures and methods to deal with instances of clauses ("abstracted" from the original clauses)
 */

#ifndef ABFORMULA_H_INCLUDED
#define ABFORMULA_H_INCLUDED

#include "idq.h"
#include "instlist.h"
#include "stack.h"

/*
 * Options:
 * 		- whether to process new instances in batches
 * 		- whether to prefer general instances to specific ones
 * 		- whether to extend the redundancy check with a time-based check
 * 		- whether to calculate varScoreSums for passive instances on the fly (only in a certain instance selection mode)
 * 		- whether to allow to mode active instances back to passive
 */

// whether to process new instances in batches
#define USE_BATCH_PROCESSING_NEW_INSTANCES
// settings for batch processing
#ifdef USE_BATCH_PROCESSING_NEW_INSTANCES
#define BATCH_SIZE_RATE 10 // batchSize = #passiveInsts / n
#define MAX_BATCH_SIZE 100 // max number of insts in batch
#endif

// whether to prefer general instances to specific ones
//#define PREFER_GENERAL_INSTANCES

// whether to extend the redundancy check with a time-based check (whether two instances have already been resolved upon the same literal before)
//#define EXTEND_REDUNDANCY_CHECK_WITH_TIME

// whether to calculate varScoreSums for passive instances on the fly (or only after each call for the SAT solver)
#define INST_VARSCORE_ON_THE_FLY

// whether to allow to move an active instance to the queue of passive instances (DOES NOT WORK, infinite execution is possible)
//#define ALLOW_TO_MOVE_ACTIVE_TO_PASSIVE

//===================================================

struct AbClause;

/*
 * Reference to a clause instance (used in the collections of passive and given instances)
 */
typedef struct InstRef {
	struct AbClause * abClause;
	unsigned int instInd; // index of the instance
} InstRef;

DECLARE_STACK(InstRef, InstRef);

/*
 * Reference to a literal of a clause instance (used in var watchers)
 */
typedef struct InstLitRef {
	struct AbClause * abClause;
	unsigned int instInd, litInd; // index of the instance and the literal
} InstLitRef;

DECLARE_STACK(InstLitRef, InstLitRef);

// give a pointer of the instance referenced by instRef
#define GET_INST_BY_ABCLAUSE_AND_INSTIND(abClause, instInd) GET_STACK((abClause)->instances, (instInd))
#define GET_INST_BY_INSTREF(instRef) GET_INST_BY_ABCLAUSE_AND_INSTIND((instRef).abClause, (instRef).instInd)

#include "bformula.h"
#include "prior_queue.h"

struct BClause;

/*
 * Clause instance
 */
typedef struct AbClauseInst {
	IList *instMask, *groundInstMask; // mask of the instance, and that of its grounding
	struct BClause * bClause; // corresponding Boolean clause

	unsigned int selLitInd; // index of the watched lit
#ifdef EXTEND_REDUNDANCY_CHECK_WITH_TIME
	unsigned int *pSelLitFrom, *pSelLitUntil; // arrays of times from/until when the corresponding lit is/was selected
#endif

	// offset of the inst in the corresponding watcher list
	// 		- if the inst is active: the offset in the watcher list of the selected literal
	// 		- if the inst is passive: the offset in the priority queue of passive insts
	unsigned int offset_instWatcher;

	char isGiven; // whether the inst is currently one of the "given" (i.e., chosen) insts

	unsigned int createdAt; // time when the inst was created

	VARSCORE_TYPE varScoreSum; // sum of scores of Boolean vars in the inst
#ifdef INST_VARSCORE_ON_THE_FLY
	unsigned int * offsets_instLitWatchers; // offsets in all the watcher lists of Boolean literals (rather, vars)
#endif
} AbClauseInst;

DECLARE_STACK(Inst, AbClauseInst);

/*
 * Clause (abstracted from input)
 */
typedef struct AbClause {
	unsigned int cntLits;
	LitID * lits;

	InstStack instances; // instances of the clause
	unsigned int cntActiveInsts; // count of active instances; active resp. passive instances are located before/after this index

	IList * depVarMask; // mask that represents dependencies of the clause (from universal vars)
} AbClause;

// whether an instRef references to a passive inst
#define IS_INST_PASSIVE(abClause,instInd) ((instInd) >= (abClause)->cntActiveInsts)

/*
 * Variable (abstracted from existential variables)
 */
typedef struct AbVariable {
	IList * depVarMask; // mask that represents dependencies of the variable (from universal vars)

	InstRefStack posLitWatchers, negLitWatchers; // literal watchers for positive/negative occurrences
} AbVariable;

/*
 * Main structure. Represents the abstracted clause set and its instances.
 */
typedef struct AbFormula {
	unsigned int cntVars;
	unsigned int cntUniversals;
	unsigned int cntExistentials;

	unsigned int cntClauses;
	unsigned int cntInstances; // overall count of instances

	AbClause * clauses;
	AbVariable * vars;

	// mapping abstract variables to original ones, and vice versa
	int * pVarMap;
	int * pExMap;
	int * pUnivMap;

	struct PriorQueue * passiveInsts; // collection of "passive" (i.e., candidate) instances
} AbFormula;

struct iDQ;
struct BFormula;
struct BClause;

void create_abFormula(struct iDQ *);

void delete_abFormula(iDQ *, AbFormula *);

void generate_passiveInsts(struct iDQ *);
unsigned int pop_passiveInsts(struct iDQ *);

#endif
