/*
 This file is part of iDQ.

 iDQ is an instantiation-based DQBF (Dependency Quantified Boolean Formula) solver.

 Andreas Froehlich, Johannes Kepler University, Linz, Austria, 2012-2014.
 Gergely Kovasznai, Vienna University of Technology, Vienna, Austria, 2012-2014.

 iDQ is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or (at
 your option) any later version.

 iDQ is distributed in the hope that it will be useful, but
 WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with iDQ.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdio.h>

#include "bformula.h"
#include "idq.h"
#include "abformula.h"
#include "mem.h"
#include "assert.h"

/*
 * Type for keys in the hashtable, representing a variable instance
 */
typedef struct HashKey {
	unsigned int varID; // ID of an existential var
	IList * instMask;
} HashKey;

/*
 * Primes for hashing
 */
static unsigned primes[] = { 111130391, 22237357, 33355519, 444476887 };

/*
 * Hash, comparison, and deletion functions for the hashtable
 */
static unsigned hash_ptr(const void * p) {
	HashKey * key = (HashKey *) p;

	IList * instMask = key->instMask;
	char isSigned = instMask->isSigned;
	unsigned res = 0, i;
	unsigned int * pUint;

	res += key->varID * primes[0];

	for (i = instMask->numOfVarsInInts, pUint = instMask->array; i; i--, pUint++) {
		res += *pUint * primes[1];
		if (isSigned)
			res += GET_SIGN_UINT_FOR_VAR_UINT(instMask, pUint) * primes[2];
	}

	return res;
}

static int cmp_ptr(const void * p, const void * q) {
	HashKey * key1 = (HashKey *) p;
	HashKey * key2 = (HashKey *) q;

	return key1->varID != key2->varID || !equal_iList(key1->instMask, key2->instMask, SIGNED);
}

static void delete_bVariable(iDQ * idq, BVariable * bVar);

static void delete_ptr(iDQ * idq, void * p) {
	HashBucket * bucket = (HashBucket *) p;

	delete_iList(((HashKey*) bucket->key)->instMask);
	delete_bVariable(idq, bucket->data.asPtr);
	DELETE(idq->mm, bucket->data.asPtr);
}

static BVariable * create_bVariable(iDQ * idq) {
	BVariable * bVar;

	CNEW(idq->mm, bVar);
#ifdef INST_VARSCORE_ON_THE_FLY
	if (USE_VARSCORE_HEURISTICS(idq))
		INIT_STACK(idq->mm, bVar->varWatchers);
#endif

	return bVar;
}

/*
 * Initializes a bFormula
 */
void create_bFormula(iDQ * idq) {
	CNEW(idq->mm, idq->bf);

	INIT_STACK(idq->mm, idq->bf->pVars);

	idq->bf->pHashTable = create_hashTable(idq, hash_ptr, cmp_ptr, delete_ptr);
}

/*
 * Deletes a Boolean var
 */
static void delete_bVariable(iDQ * idq, BVariable * bVar) {
#ifdef INST_VARSCORE_ON_THE_FLY
	if (USE_VARSCORE_HEURISTICS(idq))
		DELETE_STACK(bVar->varWatchers);
#endif
}

/*
 * Deletes a Boolean clause
 */
void delete_bClause(iDQMemMan * mm, BClause * bClause) {
	DELETEN(mm, bClause->pVars, bClause->cntLits);
}

/*
 * Deletes a BFormula
 */
void delete_bFormula(iDQ * idq, BFormula * bf) {
	BVarStack pVars = bf->pVars;
	BVariable ** pVar;
	for (pVar = pVars.start; pVar != pVars.top; pVar++)
		delete_bVariable(idq, *pVar);
	DELETE_STACK(pVars);

	delete_hashTable(bf->pHashTable);
}

/*
 * Prints the variable instances that relate to the SAT model
 */
void print_satModel(iDQ * idq) {
	AbFormula * abf = idq->abf;
	SatSolver * satSolver = idq->satSolver;

	HashBucket * b;
	HashKey * key;
	IList * instMask;
	LitID litId;
	unsigned int i, isFirstArg;

	for (b = idq->bf->pHashTable->first; b; b = b->next) {
		if (LIT_VALUE_BY_SAT_SOLVER(satSolver, b->data.asInt) < 0)
			printf("-");
		key = b->key;
		printf("%d(", abf->pExMap[key->varID]);

		instMask = key->instMask;
		isFirstArg = 1;
		for (i = 0; i < instMask->numOfVars; i++) {
			litId = getLit_iList(instMask, i + 1);

			if (!litId)
				continue;

			if (isFirstArg)
				isFirstArg = 0;
			else
				printf(",");

			if (litId < 0)
				printf("-");
			printf("%d", abf->pUnivMap[i + 1]);
		}

		printf(")\n");
	}
}

static BVariable * get_bVar_fromKey(iDQ * idq, HashKey * key) {
	HashBucket * b;

	// does it exist? if so, return pos
	if ((b = find_hashTable(idq->bf->pHashTable, key)))
		return b->data.asPtr;

	return NULL;
}

BVariable * get_bVar_fromVarInst(iDQ * idq, unsigned int varId, IList * instMask) {
	AbVariable * abVar = idq->abf->vars + varId;
	IList * litInstMask = intersection_iList(abVar->depVarMask, instMask);

	HashKey * key;
	NEW(idq->mm, key);
	key->varID = varId;
	key->instMask = union_iList(litInstMask, idq->abf->vars[varId].depVarMask); // grounding of the instance
	delete_iList(litInstMask);

	BVariable * bVar = get_bVar_fromKey(idq, key);

	delete_iList(key->instMask);
	DELETE(idq->mm, key);

	return bVar;
}

/*
 * Converts a variable instance to a Boolean variable
 */
static BVariable * convert_varInstToBVar(iDQ * idq, unsigned int varId, IList * litInstMask) {
	BVariable * bVar;

	HashKey * key;
	NEW(idq->mm, key);
	key->varID = varId;
	key->instMask = union_iList(litInstMask, idq->abf->vars[varId].depVarMask); // grounding of the instance
	delete_iList(litInstMask);

	if ((bVar = get_bVar_fromKey(idq, key))) {
		delete_iList(key->instMask);
		DELETE(idq->mm, key);
		return bVar;
	}

	HashBucket * b;
	b = insert_hashTable(idq->bf->pHashTable, key);
	return (b->data.asPtr = create_bVariable(idq));
}

/*
 * Converts a clause instance to a Boolean clause
 */
BClause * convert_instToBClause(iDQ * idq, AbClause * abClause, unsigned int instInd) {
	iDQMemMan * mm = idq->mm;
	AbFormula * abf = idq->abf;
	BFormula * bf = idq->bf;

	AbClauseInst * inst = GET_INST_BY_ABCLAUSE_AND_INSTIND(abClause, instInd);

	InstLitRef instLitRef;
#ifdef INST_VARSCORE_ON_THE_FLY
	if (USE_VARSCORE_HEURISTICS(idq)) {
		instLitRef.abClause = abClause;
		instLitRef.instInd = instInd;
	}
#endif

	BClause * bClause;
	NEW(mm, bClause);
	bClause->cntLits = abClause->cntLits;
	NEWN(mm, bClause->pVars, bClause->cntLits);

	unsigned int varId;
	BVariable * bVar;
	IList * litInstMask;
	unsigned int i;
	for (i = 0; i < abClause->cntLits; i++) {
		varId = GET_VARID(abClause->lits[i]);

		// compute the instMask of the var
		litInstMask = intersection_iList(abf->vars[varId].depVarMask, inst->instMask);

		// convert the var instance to a Boolean var
		bVar = convert_varInstToBVar(idq, varId, litInstMask);

#ifdef INST_VARSCORE_ON_THE_FLY
		// if the inst is passive, add it to the watcher list of the Boolean var
		if (USE_VARSCORE_HEURISTICS(idq) && IS_INST_PASSIVE(abClause, instInd)) {
			inst->offsets_instLitWatchers[i] = COUNT_STACK(bVar->varWatchers);
			instLitRef.litInd = i;
			PUSH_STACK(bVar->varWatchers, instLitRef);
		}
#endif

		bClause->pVars[i] = bVar;
	}

	return bClause;
}

/*
 * Increases the score of a variable
 */
static void increase_varScore(AbFormula * abf, BVariable * bVar) {
	bVar->score++;

#ifdef INST_VARSCORE_ON_THE_FLY
	InstLitRef * instLitRef;
	AbClauseInst * inst;
	for (instLitRef = bVar->varWatchers.start; instLitRef != bVar->varWatchers.top; instLitRef++) {
		inst = GET_INST_BY_ABCLAUSE_AND_INSTIND(instLitRef->abClause, instLitRef->instInd);
		inst->varScoreSum++;
		rearrangeElement_priorQueue(abf->passiveInsts, inst->offset_instWatcher);
	}
#endif
}

/*
 * Divides the VSIDS scores by a constant
 */
void decrease_varScores(iDQ * idq) {
	// divide the score of each var which has been already pushed to the SAT solver
	BVariable ** pBVar;
	unsigned int i;
	for (pBVar = idq->bf->pVars.start, i = COUNT_STACK(idq->bf->pVars); i; pBVar++, i--)
		(*pBVar)->score /= VSIDS_DIVISOR;

#ifdef INST_VARSCORE_ON_THE_FLY
	// divide the score sum of each passive inst
	AbFormula * abf = idq->abf;
	AbClause * pAbClause;
	AbClauseInst * pInst;
	for (pAbClause = abf->clauses, i = abf->cntClauses; i; pAbClause++, i--)
		for (pInst = pAbClause->instances.start + pAbClause->cntActiveInsts; pInst != pAbClause->instances.top; pInst++)
			pInst->varScoreSum /= VSIDS_DIVISOR;
#endif
}

/*
 * Adds a clause instance to the SAT solver
 */
void add_inst_toSatSolver(iDQ * idq, AbClause * abClause, unsigned int instInd) {
	BClause * bClause = GET_INST_BY_ABCLAUSE_AND_INSTIND(abClause, instInd)->bClause;
	ASSERT(bClause);

	BVariable * bVar;
	unsigned int i;
	for (i = 0; i < bClause->cntLits; i++) {
		bVar = bClause->pVars[i];
		if (!bVar->varId) {
			PUSH_STACK(idq->bf->pVars, bVar);
			bVar->varId = COUNT_STACK(idq->bf->pVars);
		}

		if (USE_VARSCORE_HEURISTICS(idq))
			// increase the score of the var
			increase_varScore(idq->abf, bVar);

		ADD_LIT_TO_SAT_SOLVER(idq->satSolver, GET_BLIT_FROM_INST(abClause, instInd, i));
	}

	ADD_LIT_TO_SAT_SOLVER(idq->satSolver, 0);

	idq->bf->cntClauses++;
}

Result getValue_lit(AbClause * abClause, unsigned int instInd, unsigned int litInd, SatSolver * satSolver) {
	return LIT_VALUE_BY_SAT_SOLVER(satSolver, GET_BLIT_FROM_INST(abClause, instInd, litInd));
}

Result getValue_inst(AbClause * abClause, unsigned int instInd, SatSolver * satSolver) {
	Result res = RESULT_UNSAT;

	BClause * bClause = GET_INST_BY_ABCLAUSE_AND_INSTIND(abClause, instInd)->bClause;
	ASSERT(bClause);

	unsigned int i;
	for (i = 0; i < abClause->cntLits; i++) {
		if (!bClause->pVars[i]->varId)
			res = RESULT_UNKNOWN;
		else if (LIT_VALUE_BY_SAT_SOLVER(satSolver, GET_BLIT_FROM_INST(abClause, instInd, i)) > 0) {
			res = RESULT_SAT;
			break;
		}
	}

	return res;
}

/*
 * Selects a satisfiable literal from a Boolean clause, and returns its index
 */
unsigned int select_satLiteral(iDQ * idq, AbClause * abClause, unsigned int instInd) {
	SatSolver * satSolver = idq->satSolver;
	BClause * bClause = GET_INST_BY_ABCLAUSE_AND_INSTIND(abClause, instInd)->bClause;
	ASSERT(bClause);

	/*#if LIT_SEL_MODE == LIT_SEL_MODE_1ST
	 unsigned int i;
	 for (i = 0; i < bClause->cntLits; i++)
	 if (LIT_VALUE_BY_SAT_SOLVER(satSolver, GET_BLIT_FROM_INST(abClause, instInd, i)) > 0)
	 return i;

	 ASSERT(0)
	 ERRORMSG("There must be a true literal in the satisfied clause");
	 return 0;*/

	if (idq->lit_sel_mode) {
		unsigned int cntLits = bClause->cntLits;
		unsigned int best_ind = cntLits;
		VARSCORE_TYPE maxScore = 0, score;

		BVariable * bVar;
		LitID bLit;
		unsigned int i;
		for (i = 0; i < cntLits; i++) {
			bVar = bClause->pVars[i];
			score = bVar->score;
			bLit = GET_BLIT_FROM_INST(abClause, instInd, i);
			if (LIT_VALUE_BY_SAT_SOLVER(satSolver, bLit) > 0 && (best_ind == cntLits || score > maxScore)) {
				best_ind = i;
				maxScore = score;
			}
		}

		ASSERT(best_ind < cntLits)
			ERRORMSG("There must be a true literal in the satisfied clause");
		return best_ind;
	} else {
		AbFormula * abf = idq->abf;

		AbClauseInst * inst = GET_INST_BY_ABCLAUSE_AND_INSTIND(abClause, instInd);

		unsigned int cntLits = abClause->cntLits;
		unsigned int best_ind = cntLits;
		LitID lit, best_lit = 0;
		unsigned int numDeps, best_numDeps = 0;
		unsigned int numSymbs, best_numSymbs = 0;

		AbVariable * abVar;
		BVariable * bVar;
		LitID bLit;
		IList * litInstMask;
		unsigned int i;
		int res;
		for (i = 0; i < cntLits; i++) {
			lit = abClause->lits[i];

			bVar = bClause->pVars[i];
			bLit = GET_BLIT_FROM_INST(abClause, instInd, i);
			if (LIT_VALUE_BY_SAT_SOLVER(satSolver, bLit) <= 0)
				continue;

			abVar = abf->vars + GET_VARID(lit);
			litInstMask = intersection_iList(abVar->depVarMask, inst->instMask);
			numSymbs = countLits_iList(litInstMask);
			numDeps = countLits_iList(abVar->depVarMask) - numSymbs;
			delete_iList(litInstMask);

			if (best_ind < cntLits) {
				// +sign
				if ((res = (lit > 0) - (best_lit > 0)) < 0)
					continue;
				// +ground
				if (!res && (res = !numDeps - !best_numDeps) < 0)
					continue;
				// -num_deps
				if (!res && (res = numDeps - best_numDeps) > 0)
					continue;
				// -num_symbs
				if (!res && (res = numSymbs - best_numSymbs) > 0)
					continue;

				if (!res)
					continue;
			}

			best_ind = i;
			best_lit = lit;
			best_numDeps = numDeps;
			best_numSymbs = numSymbs;
		}

		ASSERT(best_ind < cntLits)
			ERRORMSG("There must be a true literal in the satisfied clause");
		return best_ind;
	}
}
