/*
 This file is part of iDQ.

 iDQ is an instantiation-based DQBF (Dependency Quantified Boolean Formula) solver.

 Andreas Froehlich, Johannes Kepler University, Linz, Austria, 2014.
 Gergely Kovasznai, Vienna University of Technology, Vienna, Austria, 2014.

 iDQ is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or (at
 your option) any later version.

 iDQ is distributed in the hope that it will be useful, but
 WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with iDQ.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "hashtable.h"
#include "assert.h"

/*
 * Decide whether a hashtable should considered "full"
 */
#define IS_FULL(count, size) (count >= 0.8*size)

/*
 * Default hash, comparison, and deletion functions
 */
static unsigned hash_ptr(const void * p) {
	return 1183477 * (unsigned) (unsigned long) p;
}

static int cmp_ptr(const void * p, const void * q) {
	return ((long) p) - ((long) q);
}

static void del_ptr(iDQ * qpr, void * p) {
	DELETE(qpr->mm, p);
}

/*
 * Allocate a new hashtable with given hash, comparison, and deletion functions
 */
HashTable *
create_hashTable(iDQ * qpr, HashPtr hash, CmpPtr cmp, DeletePtr del) {
	HashTable * res;

	// allocate an empty hashtable
	CNEW(qpr->mm, res);

	res->qpr = qpr;
	res->hash = hash ? hash : hash_ptr;
	res->cmp = cmp ? cmp : cmp_ptr;
	res->del = del ? del : del_ptr;

	return res;
}

/*
 * Deallocate a hashtable
 */
void delete_hashTable(HashTable * p2iht) {
	HashBucket * p, *next;
	iDQMemMan * mm = p2iht->qpr->mm;

	for (p = p2iht->first; p; p = next) {
		next = p->next;
		p2iht->del(p2iht->qpr, p);
		DELETE(mm, p->key);
		DELETE(mm, p);
	}

	DELETEN(mm, p2iht->table, p2iht->size);
	DELETE(mm, p2iht);
}

/*
 * Enlarge a hashtable (when it is full)
 */
static void enlargeHashTable(HashTable * p2iht) {
	HashBucket * p, *chain, **old_table, **new_table;
	unsigned old_size, new_size, i, h;
	HashPtr hash;

	old_size = p2iht->size;
	old_table = p2iht->table;

	new_size = old_size ? 2 * old_size : 1;
	CNEWN(p2iht->qpr->mm, new_table, new_size);

	hash = p2iht->hash;

	for (i = 0; i < old_size; i++)
		for (p = old_table[i]; p; p = chain) {
			chain = p->chain;
			h = hash(p->key);
			h &= new_size - 1;
			p->chain = new_table[h];
			new_table[h] = p;
		}

	DELETEN(p2iht->qpr->mm, old_table, old_size);

	p2iht->size = new_size;
	p2iht->table = new_table;
}

/*
 * Find the slot which belongs to a given key
 * If there is no such slot, return NULL
 */
static HashBucket **
findPosInHashTable(HashTable * p2iht, void * key) {
	HashBucket ** p, *b;
	unsigned h;

	if (IS_FULL(p2iht->count, p2iht->size))
		enlargeHashTable(p2iht);

	ASSERT(p2iht->size > 0)
		ERRORMSG("the size of a hashtable must be positive");

	h = p2iht->hash(key);
	h &= p2iht->size - 1;

	for (p = p2iht->table + h; (b = *p) && p2iht->cmp(b->key, key); p = &b->chain)
		;

	return p;
}

HashBucket *
find_hashTable(HashTable * p2iht, void *key) {
	return *findPosInHashTable(p2iht, key);
}

/*
 * Allocate and insert a new slot for a given key
 */
HashBucket *
insert_hashTable(HashTable * p2iht, void * key) {
	HashBucket ** p, *res;

	p = findPosInHashTable(p2iht, key);
	ASSERT(!*p)
		ERRORMSG("no position found in hashtable");

	CNEW(p2iht->qpr->mm, res);
	res->key = key;
	*p = res;
	p2iht->count++;

	res->prev = p2iht->last;

	if (p2iht->first)
		p2iht->last->next = res;
	else
		p2iht->first = res;

	p2iht->last = res;

	return res;
}

/*
 * Remove the slot that belongs to a given key
 * \param stored_key_ptr a pointer to the key object in the removed slot, if stored_key_ptr is not NULL
 * \param stored_data_ptr a pointer to the data object in the removed slot, if stored_data_ptr is not NULL
 */
void remove_hashTable(HashTable * table, void * key, void ** stored_key_ptr, HashData * stored_data_ptr) {
	HashBucket ** p, *bucket;

	p = findPosInHashTable(table, key);
	bucket = *p;

	ASSERT(bucket)
		ERRORMSG("no hashbucket returned");

	*p = bucket->chain;

	if (bucket->prev)
		bucket->prev->next = bucket->next;
	else
		table->first = bucket->next;

	if (bucket->next)
		bucket->next->prev = bucket->prev;
	else
		table->last = bucket->prev;

	ASSERT(table->count > 0)
		ERRORMSG("the count of elements in hashtable must be positive before decreasing");

	table->count--;

	if (stored_key_ptr)
		*stored_key_ptr = bucket->key;

	if (stored_data_ptr)
		*stored_data_ptr = bucket->data;

	DELETE(table->qpr->mm, bucket);
}
