/*
 This file is part of iDQ.

 iDQ is an instantiation-based DQBF (Dependency Quantified Boolean Formula) solver.

 Andreas Froehlich, Johannes Kepler University, Linz, Austria, 2012-2014.
 Gergely Kovasznai, Vienna University of Technology, Vienna, Austria, 2012-2014.

 iDQ is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or (at
 your option) any later version.

 iDQ is distributed in the hope that it will be useful, but
 WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with iDQ.  If not, see <http://www.gnu.org/licenses/>.

 This file incorporates work covered by the following copyright and
 permission notice:  

   DepQBF
   Copyright 2010-2014 Florian Lonsing, Johannes Kepler University, Linz, Austria.
   Copyright 2012 Aina Niemetz, Johannes Kepler University, Linz, Austria.
*/

#include <stdio.h>
#include <string.h>
#include <time.h>

#include "idq.h"
#include "mem.h"
#include "abformula.h"
#include "qbformula.h"
#include "bformula.h"
#include "assert.h"
#include "preproc.h"
#include "config.h"

/*
 * Variable for measuring time
 */
double tSatSolve;
double tTotalTime;
double tInstantiate;
double tBasicSetup;
clock_t clStart;
clock_t clEnd;

/*
 * Creates a solver instance
 */
iDQ * idq_create() {
	iDQMemMan * mm = idq_create_mem_man();

	iDQ * idq;
	CNEW(mm, idq);

	idq->mm = mm;

	NEW(mm, idq->qbf);

	idq->satSolver = INIT_SAT_SOLVER();

	INIT_STACK(mm, idq->tempLitStack);

	return idq;
}

/*
 * Deletes a Clause
 */
void delete_clause(iDQMemMan * mm, Clause * clause) {
	DELETEN(mm, clause->lits, clause->cntLits);
}

/*
 * Deletes a Variable
 */
void delete_variable(iDQMemMan * mm, Variable * var) {
	DELETEN(mm, var->dependencies, var->cntDependencies);
}

/*
 * Deletes a solver instance
 */
void idq_delete(iDQ * idq) {
	iDQMemMan * mm = idq->mm;

	RELEASE_SAT_SOLVER(idq->satSolver);

	delete_qbFormula(mm, idq->qbf);
	DELETE(mm, idq->qbf);

	if (idq->abf) {
		delete_abFormula(idq, idq->abf);
		DELETE(mm, idq->abf);
	}

	if (idq->bf) {
		delete_bFormula(idq, idq->bf);
		DELETE(mm, idq->bf);
	}

	DELETE(mm, idq);

	idq_delete_mem_man(mm);
}

/*
 * Reduces variable dependencies by calling DepQBF
 */
static void reduce_deps(QDPLL * qdpll, QBFormula * qbf) {
	ASSERT(qdpll);
	ASSERT(qbf);

	unsigned int maxVarId = qdpll_get_max_declared_var_id(qdpll);
	Variable * var;
	unsigned int *deps, *pDep;
	unsigned int i, j, k;

	qdpll_init_deps(qdpll);

	for (i = 1; i <= maxVarId; i++) {
		if (!qdpll_is_var_declared(qdpll, i))
			continue;

		if (qdpll_get_scope_type(qdpll, qdpll_get_nesting_of_var(qdpll, i)) == QDPLL_QTYPE_FORALL)
			continue;

		var = qbf->pVars + i;
		deps = var->dependencies;

		for (pDep = (deps = var->dependencies), j = 0; j < var->cntDependencies; j++) {
			if (qdpll_is_var_declared(qdpll, deps[j]) && qdpll_var_depends(qdpll, deps[j], i))
				continue;

			var->cntDependencies--;
			for (k = j; k < var->cntDependencies; k++)
				deps[k] = deps[k + 1];
			j--;
		}
	}
}

/*
 * Parses command line arguments
 */
static int parse_cmdline_args(iDQ * idq, int argc, char **argv) {
	char ** pArg;
	for (pArg = argv + 1, argc--; argc; pArg++, argc--) {
		if (!strcmp(*pArg, "-h") || !strcmp(*pArg, "--help")) {
			printf("usage: idq [<option>...] <(D)QDIMACS file>\n\n");
			printf("where <option> is one of the following:\n\n");
			printf("  -h|--help                        print usage information and exit\n");
			printf("  --copyright                      print copyright and exit\n");
			printf("  --version                        print version and exit\n");
			printf("\n");
			printf("  --preproc=[on|off]               preprocessing is on/off (default: off)\n");
			printf("  --heur=[default|vsids]           heuristics for instance+literal selection\n");
			printf("  -v|--verbose                     increase verbosity (default: 0)\n");
			printf("  --model                          print satisfying model\n");
			printf("  -d|--dep-analysis                analyze dependencies if input is QDIMACS\n");
			printf("  --dump-dqdimacs <file>           dump QDIMACS file\n");
			printf("  --dump-tptp <file>               dump TPTP file\n");
			return 1;
		}
		if (!strcmp(*pArg, "--copyright")) {
			printf("Copyright (C) 2013-2014 Andreas Fröhlich, Gergely Kovásznai\n");
			printf("Institute for Formal Models and Verification\n");
			printf("Johannes Kepler University, Linz, Austria\n");
			return 1;
		}
		if (!strcmp(*pArg, "--version")) {
			printf("%s\n", IDQ_VERSION);
			return 1;
		}
		if (!strcmp(*pArg, "--preproc=on")) {
			idq->preproc = 1;
			continue;
		}
		if (!strcmp(*pArg, "--preproc=off")) {
			idq->preproc = 0;
			continue;
		}
		if (!strcmp(*pArg, "--heur=default")) {
			idq->inst_sel_mode = idq->lit_sel_mode = idq->var_score_mode = 0;
			continue;
		}
		if (!strcmp(*pArg, "--heur=vsids")) {
			idq->inst_sel_mode = idq->lit_sel_mode = idq->var_score_mode = 1;
			continue;
		}
		if (!strcmp(*pArg, "-v")) {
			idq->verbosity_level++;
			continue;
		}
		if (!strcmp(*pArg, "--model")) {
			idq->print_model = 1;
			continue;
		}
		if (!strcmp(*pArg, "-d") || !strcmp(*pArg, "--dep-analysis")) {
			if (!idq->qdpll)
				idq->qdpll = qdpll_create();
			continue;
		}
		if (!strcmp(*pArg, "--dump-dqdimacs")) {
			idq->dump_dqdimacs = 1;
			pArg++;
			argc--;
			ASSERT(argc)
				ERRORMSG("no file name after '--dump-dimacs'");
			ASSERT((idq->outputFile = fopen(*pArg, "w")))
				ERRORMSG("cannot create file '%s'", *pArg);
			continue;
		}
		if (!strcmp(*pArg, "--dump-tptp")) {
			idq->dump_tptp = 1;
			pArg++;
			argc--;
			ASSERT(argc)
				ERRORMSG("no file name after '--dump-tptp'");
			ASSERT((idq->outputFile = fopen(*pArg, "w")))
				ERRORMSG("cannot create file '%s'", *pArg);
			continue;
		}

		ASSERT(!idq->inputFile)
			ERRORMSG("at most 1 input file allowed");
		ASSERT((idq->inputFile = fopen(*pArg, "r")))
			ERRORMSG("cannot open file '%s'", *pArg);
	}

	ASSERT(idq->inputFile)
		ERRORMSG("no input file specified");

	return 0;
}

/*
 * Main method for solving the input problem
 */
Result idq_solve(iDQ * idq) {
	AbFormula * abf = idq->abf;
	BFormula * bf = idq->bf;
	SatSolver * satSolver = idq->satSolver;

	Result res = RESULT_UNKNOWN;

	tSatSolve = 0;
	tInstantiate = 0;
	double tCurrSolve;

	unsigned int cntNewInstances;

	// loop until SAT solver returns UNSAT or no passive instance remains (=SAT)
	do {
		if (idq->clock) {
			// read the given (i.e., chosen) instances from the collection of passive instances
			if (!(cntNewInstances = pop_passiveInsts(idq))) // if there is no new instance, the result is SAT
				break;
		}

		idq->clock++;
		if (idq->verbosity_level > 0 && idq->clock % 10 == 1)
			printf("Starting SAT solver in iteration %d...\n", idq->clock);

		// call the SAT solver
		clStart = clock();
		res = CHECK_SAT_WITH_SAT_SOLVER(satSolver);
		ASSERT(res == RESULT_SAT || res == RESULT_UNSAT);
		clEnd = clock();
		tCurrSolve = (double) (clEnd - clStart) / CLOCKS_PER_SEC;
		tSatSolve += tCurrSolve;

		// if the SAT solver returns UNSAT
		if (res == RESULT_UNSAT)
			return RESULT_UNSAT;

		// generate new passive instances
		clStart = clock();
		generate_passiveInsts(idq);
		clEnd = clock();
		tInstantiate += (double) (clEnd - clStart) / CLOCKS_PER_SEC;
	} while (abf->passiveInsts->count);

	if (idq->print_model)
		print_satModel(idq);

	return RESULT_SAT;
}

/*
 * MAIN
 */
int main(int argc, char **argv) {
	iDQ * idq = idq_create();

	// parse command line arguments
	if (parse_cmdline_args(idq, argc, argv)) {
		idq_delete(idq);
		return 0;
	}

	if (idq->verbosity_level > 0) {
		printf("Preprocessing: %s\n", idq->preproc ? "on" : "off");
		printf("Instance selection mode: %s\n", idq->inst_sel_mode ? "maximal var score" : "default");
		printf("Literal selection mode: %s\n", idq->lit_sel_mode ? "maximal var score" : "default");
		if (USE_VARSCORE_HEURISTICS(idq))
			printf("Variable score more : %s\n", idq->var_score_mode ? "vsids" : "count of occurrences");
	}

	clock_t totalStart = clock();
	clStart = clock();

	Result res = RESULT_UNKNOWN;

	// parse input DQDIMACS file
	if (idq->verbosity_level > 0)
		printf("Parsing input file...");
	parse_dqdimacs(idq);
	fclose(idq->inputFile);
	if (idq->verbosity_level > 0)
		printf(" OK\n");

	QBFormula * qbf = idq->qbf;

	// reduce dependencies
	if (idq->qdpll) {
		if (idq->verbosity_level > 0)
			printf("Reducing dependencies with DepQBF...");
		reduce_deps(idq->qdpll, qbf);
		qdpll_delete(idq->qdpll);
		if (idq->verbosity_level > 0)
			printf(" OK\n");
	}

	if (idq->preproc)
		blocked_clause_elimination(idq);

	if (idq->dump_dqdimacs) {
		if (idq->verbosity_level > 0)
			printf("Dumping DQDIMACS...");
		dump_dqdimacs(idq);
		if (idq->verbosity_level > 0)
			printf(" OK\n");

		idq_delete(idq);
		return 0;
	}

	// initialize additional data structures
	create_bFormula(idq);
	create_abFormula(idq);

	if (idq->dump_tptp) {
		if (idq->verbosity_level > 0)
			printf("Dumping TPTP...");
		dump_tptp(idq);
		if (idq->verbosity_level > 0)
			printf(" OK\n");

		idq_delete(idq);
		return 0;
	}

	clEnd = clock();
	tBasicSetup = (double) (clEnd - clStart) / CLOCKS_PER_SEC;

	// solve the input problem
	if (idq->verbosity_level > 0)
		printf("Starting DQBF solving\n");
	res = idq_solve(idq);
	if (idq->verbosity_level > 0)
		printf("DQBF solving finished\n");

	clock_t totalEnd = clock();
	tTotalTime = (double) (totalEnd - totalStart) / CLOCKS_PER_SEC;

	/*	unsigned int fullExp, i;
	 for (i = 0, fullExp = 0; i < abf->cntClauses; i++) {
	 unsigned int bla = 1 << instListCountLits(abf->pClauses[i].depVarMask);

	 //printf("%d ", bla);
	 fullExp += bla;
	 }*/

	// print statistics
	printf("Number of universal variables: %d\n", qbf->cntUniversals);
	printf("Number of original clauses: %d\n", qbf->cntClauses);
	printf("Number of instances: %d\n", idq->abf->cntInstances);
	printf("Number of Boolean instances: %d\n", idq->bf->cntClauses);
//	printf("Size of full expansion: %d\n\n", fullExp);
	printf("Number of SAT solver iterations: %d\n", idq->clock);

	printf("TotalTime: %f\n", tTotalTime);
	printf("Time for basic setup: %f\n", tBasicSetup);
	printf("Time for SAT solving: %f\n", tSatSolve);
	printf("Time for instantiation: %f\n", tInstantiate);

	if (res == RESULT_SAT) {
		fprintf(stdout, "SAT\n");
	} else if (res == RESULT_UNSAT)
		fprintf(stdout, "UNSAT\n");
	else {
		ASSERT(res == RESULT_UNKNOWN);
		fprintf(stdout, "UNKNOWN\n");
	}

	idq_delete(idq);

	return res;
}
