/*
 This file is part of iDQ.

 iDQ is an instantiation-based DQBF (Dependency Quantified Boolean Formula) solver.

 Andreas Froehlich, Johannes Kepler University, Linz, Austria, 2012-2014.
 Gergely Kovasznai, Vienna University of Technology, Vienna, Austria, 2012-2014.

 iDQ is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or (at
 your option) any later version.

 iDQ is distributed in the hope that it will be useful, but
 WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with iDQ.  If not, see <http://www.gnu.org/licenses/>.

 This file incorporates work covered by the following copyright and
 permission notice:  

   DepQBF
   Copyright 2010-2014 Florian Lonsing, Johannes Kepler University, Linz, Austria.
   Copyright 2012 Aina Niemetz, Johannes Kepler University, Linz, Austria.
*/

#ifndef IDQ_H_INCLUDED
#define IDQ_H_INCLUDED

#include <stdio.h>
#include <limits.h>

#include "mem.h"
#include "qdpll.h"

/*
 * Options (can be set from command line):
 * 		- which (passive) instance selection mode to use
 * 		- which literal selection mode to use
 * 		- which type of var score to use
 * 	See more options in abformula.h and bformula.h!
 */

#define USE_VARSCORE_HEURISTICS(idq) ((idq)->inst_sel_mode || (idq)->lit_sel_mode)

//===================================================

#include "stack.h"
#include "satsolver.h"

struct QBFormula;
struct BFormula;
struct AbFormula;

//typedef int LitID;
//typedef unsigned int VarID;
#define GET_VARID(litID) (litID>0?litID:-litID)
#define INVALID_LIT_IND UINT_MAX

#include "instlist.h"

typedef enum Result {
	RESULT_UNKNOWN = 0, RESULT_SAT = 10, RESULT_UNSAT = 20
} Result;

typedef enum QuantifierType {
//	QTYPE_EXISTS = -1, QTYPE_UNDEF = 0, QTYPE_FORALL = 1
	QTYPE_EXISTS = 0, QTYPE_FORALL = 1
} QuantifierType;

typedef struct Variable {
	QuantifierType quantifier;
	unsigned int cntDependencies;
	unsigned int * dependencies;
} Variable;

/*
 * Clause
 */
typedef struct Clause {
	unsigned int cntLits;
	LitID * lits;
	int is_deleted :1;
} Clause;

DECLARE_STACK(LitId, LitID);

/*
 * Main data structure of this solver
 */
typedef struct iDQ {
	iDQMemMan * mm;
	unsigned int clock; // number of solving iterations so far
	FILE *inputFile, *outputFile; // input and output files
	QDPLL * qdpll; // DepQBF instance
	struct QBFormula * qbf; // original clause set
	struct AbFormula * abf; // abstracted clause set
	struct BFormula * bf; // Boolean clause set
	SatSolver * satSolver; // SAT solver instance
	LitIdStack tempLitStack; // temporary stack for literal ids

	char verbosity_level;
	int print_model :1; // flag for printing satisfying model
	int dump_dqdimacs :1; // flag for dumping DQDIMACS
	int dump_tptp :1; // flag for dumping TPTP

	int preproc :1; // flag for preprocessing

	int inst_sel_mode :1; // instance selection mode (0=default, 1=max var scores)
	int lit_sel_mode :1; // literal selection mode (0=default, 1=max var scores)
	int var_score_mode :1; // variable score mode (0=count of occurrences, 1=vsids)
} iDQ;

iDQ * idq_create();

void delete_clause(iDQMemMan * mm, Clause * clause);
void delete_variable(iDQMemMan * mm, Variable * var);
void idq_delete();

Result idq_solve();

#endif
