/*
 This file is part of iDQ.

 iDQ is an instantiation-based DQBF (Dependency Quantified Boolean Formula) solver.

 Andreas Froehlich, Johannes Kepler University, Linz, Austria, 2012-2014.
 Gergely Kovasznai, Vienna University of Technology, Vienna, Austria, 2012-2014.

 iDQ is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or (at
 your option) any later version.

 iDQ is distributed in the hope that it will be useful, but
 WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with iDQ.  If not, see <http://www.gnu.org/licenses/>.
*/

/*
 * instList.c
 *
 *  Created on: Oct 11, 2012
 *      Author: kovasz
 */

#include <stdlib.h>
#include <math.h>

#include "instlist.h"
#include "assert.h"

#define MIN(X,Y) ((X) < (Y) ? (X) : (Y))
#define MAX(X,Y) ((X) > (Y) ? (X) : (Y))

// how many bits an int contains
#define numOfBitsInInt (sizeof(int) << 3)
#define numOfInts(numOfBits) (((numOfBits - 1) / numOfBitsInInt) + 1)

#define GET_VAR_UINT(iList, index) iList->array[index]
#define GET_SIGN_UINT(iList, index) iList->array[iList->numOfVarsInInts + index]
//#define SET_VAR_UINT(iList, index, value) GET_VAR_UINT(iList, index) = value
//#define SET_SIGN_UINT(iList, index, value) GET_SIGN_UINT(iList, index) = value

#define HAVE_SIGN_CONFLICT(iList1, pUint1, iList2, pUint2) \
	(iList1->isSigned && iList2->isSigned && (GET_SIGN_UINT_FOR_VAR_UINT(iList1, pUint1) & (*pUint1 & *pUint2)) != (GET_SIGN_UINT_FOR_VAR_UINT(iList2, pUint2) & (*pUint1 & *pUint2)))

IList* create_iList(iDQMemMan * mem, unsigned int numOfVars, char isSigned) {
//	ASSERT(numOfVars > 0);

	IList* newInstList;

	NEW(mem, newInstList);

	newInstList->mem = mem;

	newInstList->numOfVars = numOfVars;
	newInstList->numOfVarsInInts = numOfVars ? numOfInts(numOfVars) : 0;

	newInstList->isSigned = isSigned;

	if (numOfVars)
		CNEWN(mem, newInstList->array, newInstList->numOfVarsInInts << (isSigned ? 1 : 0));
	else
		newInstList->array = NULL;

	return newInstList;
}

void delete_iList(void* _iList) {
	ASSERT(_iList);

	IList* iList = (IList*) _iList;

	if (iList->array)
		DELETEN(iList->mem, iList->array, iList->numOfVarsInInts << (iList->isSigned ? 1 : 0));

	DELETE(iList->mem, iList);
}

void print_iList(IList* iList, FILE* out) {
	ASSERT(iList);

	unsigned numOfVarsInInt = numOfBitsInInt;
	unsigned int varUint, signUint = 0;
	unsigned int i, j;
	char isFirstLit = 1;
	for (i = 0; i < iList->numOfVarsInInts; i++) {
		varUint = GET_VAR_UINT(iList, i);
		if (iList->isSigned)
			signUint = GET_SIGN_UINT(iList, i);

		for (j = 0; varUint; j++) {
			if (varUint & 1) {
				printf("%s%s%d", isFirstLit ? "" : ",", (iList->isSigned && (signUint & 1)) ? "-" : "", i * numOfVarsInInt + j + 1);
				isFirstLit = 0;
			}

			varUint >>= 1;
			if (iList->isSigned)
				signUint >>= 1;
		}
	}
}

/*
 * Given a VarID, returns its absence (0) or the sign (-1 or 1) in the IList
 */
LitID getLit_iList(IList* iList, VarID varId) {
	ASSERT(iList);

	varId--;

	ASSERT(varId < iList->numOfVars);

	//varId <<= 1;

	// calculate the intIndex
	unsigned int intIndex = varId / numOfBitsInInt;

	ASSERT(intIndex < iList->numOfVarsInInts);

	// calculate the bitIndex within the given uint
	varId %= numOfBitsInInt;

	// get the intIndexth uint, and get its bitIndexth bit
	if ((GET_VAR_UINT(iList, intIndex)>> varId) & 1) {
		if (iList->isSigned && ((GET_SIGN_UINT(iList, intIndex) >> varId) & 1))
		return -1;
		else
		return 1;
	}

	return 0;
}

/*
 * Sets a given LitID in the IList
 */
void setLit_iList(IList* iList, LitID litId) {
	ASSERT(iList);

	VarID varId = GET_VARID(litId) - 1;

	ASSERT(varId < iList->numOfVars);

	// calculate the intIndex
	unsigned int intIndex = varId / numOfBitsInInt;

	ASSERT(intIndex < iList->numOfVarsInInts);

	// calculate the bitIndex within the given uint
	varId %= numOfBitsInInt;

	// set the bitIndexth bit of the intIndexth uint to 1
	GET_VAR_UINT(iList, intIndex)|= 1 << varId;
	if (iList->isSigned && litId < 0)
		GET_SIGN_UINT(iList, intIndex)|= 1 << varId;
	}

	/*
	 * Intersects an IList with another, and overwrites the first one with the result
	 */
void intersectWith_iList(IList* iList1, IList* iList2) {
	ASSERT(iList1 && iList2);

	unsigned int numOfVarsInInts = MIN(iList1->numOfVarsInInts, iList2->numOfVarsInInts);
	char checkSigns = (iList1->isSigned || iList2->isSigned);

	unsigned int *pUint1, *pUint2;
	unsigned int i;
	for (i = numOfVarsInInts, pUint1 = iList1->array, pUint2 = iList2->array; i; i--, pUint1++, pUint2++) {
		*pUint1 &= *pUint2;

		// check whether the common vars have the same signs
		assert(!checkSigns || !HAVE_SIGN_CONFLICT(iList1, pUint1, iList2, pUint2));

		if (iList1->isSigned)
			GET_SIGN_UINT_FOR_VAR_UINT(iList1, pUint1) &= *pUint1;
	}
}

/*
 * Returns the intersections of two ILists
 */
IList * intersection_iList(IList* iList1, IList* iList2) {
	ASSERT(iList1 && iList2);

	IList * result = create_iList(iList1->mem, MIN(iList1->numOfVars, iList2->numOfVars), (iList1->isSigned || iList2->isSigned));

	unsigned int *pUint1, *pUint2, *pResultUint;
	unsigned int i;
	for (i = result->numOfVarsInInts, pUint1 = iList1->array, pUint2 = iList2->array, pResultUint = result->array; i; i--, pUint1++, pUint2++, pResultUint++) {
		*pResultUint = *pUint1 & *pUint2;

		if (!result->isSigned)
			continue;

		// check whether the common vars have the same signs
		if (HAVE_SIGN_CONFLICT(iList1, pUint1, iList2, pUint2)) {
			delete_iList(result);
			return NULL;
		}

		if (iList1->isSigned)
			GET_SIGN_UINT_FOR_VAR_UINT(result, pResultUint) = (GET_SIGN_UINT_FOR_VAR_UINT(iList1, pUint1) & *pResultUint);
		else if (iList2->isSigned)
			GET_SIGN_UINT_FOR_VAR_UINT(result, pResultUint) = (GET_SIGN_UINT_FOR_VAR_UINT(iList2, pUint2) & *pResultUint);
	}

	return result;
}

/*
 * Returns the union of two ILists
 */
IList * union_iList(IList* iList1, IList* iList2) {
	ASSERT(iList1 && iList2);

	IList * result = create_iList(iList1->mem, MAX(iList1->numOfVars, iList2->numOfVars), (iList1->isSigned || iList2->isSigned));

	unsigned int *pUint1, *pUint2, *pResultUint;
	unsigned int i;
	for (i = result->numOfVarsInInts, pUint1 = iList1->array, pUint2 = iList2->array, pResultUint = result->array; i; i--, pUint1++, pUint2++, pResultUint++) {
		*pResultUint = *pUint1 | *pUint2;

		if (!result->isSigned)
			continue;

		// check whether the common vars have the same signs
		if (HAVE_SIGN_CONFLICT(iList1, pUint1, iList2, pUint2)) {
			delete_iList(result);
			return NULL;
		}

		GET_SIGN_UINT_FOR_VAR_UINT(result, pResultUint) = (iList1->isSigned ? GET_SIGN_UINT_FOR_VAR_UINT(iList1, pUint1) : 0) | (iList2->isSigned ? GET_SIGN_UINT_FOR_VAR_UINT(iList2, pUint2) : 0);
	}

	return result;
}

/*
 * Whether two ILists are equal, with or without considering the signs of literals
 */
char equal_iList(IList* iList1, IList* iList2, char isSigned) {
	ASSERT(iList1 && iList2);

	unsigned int *pUint1, *pUint2;
	unsigned int i;
	for (i = iList1->numOfVarsInInts, pUint1 = iList1->array, pUint2 = iList2->array; i; i--, pUint1++, pUint2++)
		if (*pUint1 != *pUint2 || (isSigned && HAVE_SIGN_CONFLICT(iList1, pUint1, iList2, pUint2)))
			return 0;

	return 1;
}

/*
 * Whether an IList is a subset of another, with or without considering the signs of literals
 */
char subseteq_iList(IList* iList1, IList* iList2, char isSigned) {
	ASSERT(iList1 && iList2);

	unsigned int *pUint1, *pUint2;
	unsigned int i;
	for (i = iList1->numOfVarsInInts, pUint1 = iList1->array, pUint2 = iList2->array; i; i--, pUint1++, pUint2++) {
		// if *pUint1 is a subset of *pUint2
		if ((*pUint1 | *pUint2) != *pUint2 || (isSigned && HAVE_SIGN_CONFLICT(iList1, pUint1, iList2, pUint2)))
			return 0;
	}

	return 1;
}

/*
 * Returns the number of literals
 */
unsigned int countLits_iList(IList* iList) {
	assert(iList);

	unsigned int cntOnes, actUint, *pUint, i;

	for (i = iList->numOfVarsInInts, pUint = iList->array, cntOnes = 0; i; i--, pUint++)
		for (actUint = *pUint; actUint; actUint >>= 1)
			cntOnes += (actUint & 1);

	return cntOnes;
}

/*void iListOverrideWith(IList * iList1, IList* iList2) {
 ASSERT(iList1 && iList2);

 unsigned int *pUint1, *pUint2;
 unsigned int intersectionUint;
 unsigned int i;
 for (i = iList1->numOfVarsInInts, pUint1 = iList1->array, pUint2 = iList2->array; i; i--, pUint1++, pUint2++) {
 intersectionUint = *pUint1 & *pUint2;

 *pUint1 |= *pUint2;

 if (iList1->isSigned)
 GET_SIGN_UINT_FOR_VAR_UINT(iList1, pUint1) = (GET_SIGN_UINT_FOR_VAR_UINT(iList1, pUint1) & ~intersectionUint) | GET_SIGN_UINT_FOR_VAR_UINT(iList2, pUint2);
 }
 }*/
