/*
 This file is part of iDQ.

 iDQ is an instantiation-based DQBF (Dependency Quantified Boolean Formula) solver.

 Andreas Froehlich, Johannes Kepler University, Linz, Austria, 2012-2014.
 Gergely Kovasznai, Vienna University of Technology, Vienna, Austria, 2012-2014.

 iDQ is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or (at
 your option) any later version.

 iDQ is distributed in the hope that it will be useful, but
 WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with iDQ.  If not, see <http://www.gnu.org/licenses/>.

 This file incorporates work covered by the following copyright and
 permission notice:  

   DepQBF
   Copyright 2010-2014 Florian Lonsing, Johannes Kepler University, Linz, Austria.
   Copyright 2012 Aina Niemetz, Johannes Kepler University, Linz, Austria.
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stddef.h>

#include "mem.h"
#include "idq.h"
#include "bformula.h"
#include "assert.h"

#define ABORT_MEM(cond,msg) \
  do { \
    if (cond) \
      { \
        fprintf (stderr, "[qpr_mem] %s: %s\n", __func__, msg); \
	fflush (stderr); \
        abort ();	 \
      } \
  } while (0)

iDQMemMan *
idq_create_mem_man() {
	iDQMemMan *mm = (iDQMemMan *) malloc(sizeof(iDQMemMan));
	ABORT_MEM(!mm, "could not allocate memory!");
	memset(mm, 0, sizeof(iDQMemMan));
	return mm;
}

void idq_delete_mem_man(iDQMemMan * mm) {
	ABORT_MEM(!mm, "null pointer encountered!");

/*	ASSERT(!mm->cur_allocated)
		ERRORMSG("%d bytes have not been deallocated, out of %d", mm->cur_allocated, mm->max_allocated);*/

	free(mm);
}

void *
idq_malloc(iDQMemMan * mm, size_t size) {
	/* Mem-limit is given in MB. */
	if (mm->limit && mm->limit < (mm->cur_allocated + size) / 1024 / 1024) {
		fprintf(stderr, "Attempted to allocate total %f MB (limit = %lu MB)\n", ((unsigned long) (mm->cur_allocated + size)) / 1024 / (float) 1024, (unsigned long) mm->limit);
		ABORT_MEM(1, "mem-limit exceeded!");
	}
	void *r = malloc(size);
	ABORT_MEM(!r, "could not allocate memory!");
	memset(r, 0, size);
	mm->cur_allocated += size;
	if (mm->cur_allocated > mm->max_allocated)
		mm->max_allocated = mm->cur_allocated;
	return r;
}

void *
idq_calloc(iDQMemMan * mm, size_t nobj, size_t size) {
	/* Mem-limit is given in MB. */
	if (mm->limit && mm->limit < (mm->cur_allocated + nobj * size) / 1024 / 1024) {
		fprintf(stderr, "Attempted to allocate total %f MB (limit = %lu MB)\n", ((unsigned long) (mm->cur_allocated + nobj * size)) / 1024 / (float) 1024, (unsigned long) mm->limit);
		ABORT_MEM(1, "mem-limit exceeded!");
	}
	void *r = calloc(nobj, size);
	ABORT_MEM(!r, "could not allocate memory!");
	mm->cur_allocated += nobj * size;
	if (mm->cur_allocated > mm->max_allocated)
		mm->max_allocated = mm->cur_allocated;
	return r;
}

void *
idq_realloc(iDQMemMan * mm, void *ptr, size_t old_size, size_t new_size) {
	ptr = realloc(ptr, new_size);
	ABORT_MEM(!ptr, "could not allocate memory!");
	if (new_size > old_size)
		memset(((char *) ptr) + old_size, 0, new_size - old_size);
	mm->cur_allocated -= old_size;
	mm->cur_allocated += new_size;
	if (mm->cur_allocated > mm->max_allocated)
		mm->max_allocated = mm->cur_allocated;
	return ptr;
}

void idq_free(iDQMemMan * mm, void *ptr, size_t size) {
	ABORT_MEM(!mm, "null pointer encountered!");
	free(ptr);
	ASSERT(mm->cur_allocated >= size);
	mm->cur_allocated -= size;
}

size_t idq_max_allocated(iDQMemMan * mm) {
	return mm->max_allocated;
}

size_t idq_cur_allocated(iDQMemMan * mm) {
	return mm->cur_allocated;
}

void idq_set_mem_limit(iDQMemMan * mm, size_t limit) {
	ABORT_MEM(limit <= 0, "mem-limit must be greater than 0!");
	mm->limit = limit;
}

size_t idq_get_mem_limit(iDQMemMan * mm) {
	return mm->limit;
}
