/*
 This file is part of iDQ.

 iDQ is an instantiation-based DQBF (Dependency Quantified Boolean Formula) solver.

 Andreas Froehlich, Johannes Kepler University, Linz, Austria, 2012-2014.
 Gergely Kovasznai, Vienna University of Technology, Vienna, Austria, 2012-2014.

 iDQ is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or (at
 your option) any later version.

 iDQ is distributed in the hope that it will be useful, but
 WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with iDQ.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "preproc.h"

#include "qbformula.h"
#include "assert.h"

static char depends_on(QBFormula * qbf, Variable * eVar, unsigned int varId) {
	Variable * var = qbf->pVars + varId;
	unsigned int dep, i, j;
	char found;

	if (var->quantifier == QTYPE_FORALL) {
		for (i = 0; i < eVar->cntDependencies; i++) {
			if (eVar->dependencies[i] == varId)
				return 1;
		}
		return 0;
	}

	for (i = 0; i < var->cntDependencies; i++) {
		dep = var->dependencies[i];
		found = 0;
		for (j = 0; !found && j < eVar->cntDependencies; j++) {
			if (dep == eVar->dependencies[j])
				found = 1;
		}
		if (!found)
			return 0;
	}
	return 1;
}

static char is_resolvent_tautology(QBFormula * qbf, Clause * clause1, Clause * clause2, Variable * eVar) {
	LitID lit;
	VarID varId;
	Variable * var;
	unsigned int i, j;

	for (i = 0; i < clause1->cntLits; i++) {
		lit = clause1->lits[i];
		varId = GET_VARID(lit);
		if (eVar == qbf->pVars + varId)
			continue;

		if (!depends_on(qbf, eVar, varId))
			continue;

		for (j = 0; j < clause2->cntLits; j++)
			if (lit == -clause2->lits[j])
				return 1;
	}

	return 0;
}

static char is_blocking_literal(QBFormula * qbf, Clause * clause, LitID eLit) {
	Clause * clause2;
	VarID varId;
	Variable * var;
	unsigned int cntClauses = qbf->cntClauses, i, j;

	for (i = 0; i < cntClauses; i++) {
		clause2 = qbf->pClauses + i;
		if (clause == clause2 || clause2->is_deleted)
			continue;

		for (j = 0; j < clause2->cntLits; j++)
			if (eLit == -clause2->lits[j])
				break;

		if (j == clause2->cntLits)
			continue;

		if (!is_resolvent_tautology(qbf, clause, clause2, qbf->pVars + GET_VARID(eLit)))
			return 0;
	}

	return 1;
}

void blocked_clause_elimination(iDQ * idq) {
	ASSERT(idq->qbf);

	unsigned int cnt_deleted = 0;
	char deleted = 1;

	QBFormula * qbf = idq->qbf;
	Clause * clause;
	LitID lit;
	unsigned int cntClauses = qbf->cntClauses, i, j;

	while (deleted) {
		deleted = 0;

		for (i = 0; i < cntClauses; i++) {
			clause = qbf->pClauses + i;
			if (clause->is_deleted)
				continue;

			for (j = 0; j < clause->cntLits; j++) {
				lit = clause->lits[j];

				if (qbf->pVars[GET_VARID(lit)].quantifier == QTYPE_FORALL)
					continue;

				if (is_blocking_literal(qbf, clause, lit)) {
					deleted = clause->is_deleted = 1;
					cnt_deleted++;
					break;
				}
			}
		}

	}

	qbf->cntDeletedClauses = cnt_deleted;
	if (idq->verbosity_level > 0)
		printf("Blocked clause elimination deleted %d clause(s)\n", cnt_deleted);
}
