/*
 This file is part of iDQ.

 iDQ is an instantiation-based DQBF (Dependency Quantified Boolean Formula) solver.

 Andreas Froehlich, Johannes Kepler University, Linz, Austria, 2012-2014.
 Gergely Kovasznai, Vienna University of Technology, Vienna, Austria, 2012-2014.

 iDQ is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or (at
 your option) any later version.

 iDQ is distributed in the hope that it will be useful, but
 WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with iDQ.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "prior_queue.h"
#include "assert.h"

#include <stdlib.h>

// a binary tree stored in an array with the root at 0 and children at indices i*2 resp. i*2+1
#define PARENT_INDEX(i) ((i+1)/2 - 1)
#define LEFT_CHILD_INDEX(i) ((i+1)*2 - 1)
#define RIGHT_CHILD_INDEX(i) ((i+1)*2)

static void del(iDQMemMan * mem, PassiveInstRef * a) {
	DELETE(mem, a);
}

/*
 * Allocate a new priority queue with given comparison and deletion functions
 */
PriorQueue *create_priorQueue(iDQMemMan *mem, AbFormula *abf, CmpPtr cmp) {
	PriorQueue * res;

	// allocate an empty priority queue
	CNEW(mem, res);

	res->mem = mem;
	res->abf = abf;

	res->cmp = cmp;

	return res;
}

// updates the watcher offset of a passive instance
#define UPDATE_WATCHER_OFFSETS(pqueue, oldInd, newInd) GET_INST_BY_INSTREF((pqueue)->list[(oldInd)]->instRef)->offset_instWatcher = (newInd)

/*
 * Enqueue a new element
 */
unsigned int enqueue_priorQueue(PriorQueue *pqueue, PassiveInstRef *data) {
	if (pqueue->count == pqueue->size) {
		pqueue->size = pqueue->size ? 2 * pqueue->size : 1;
		REALLOC(pqueue->mem, pqueue->list, pqueue->count, pqueue->size);
	}

	unsigned i = pqueue->count++;
	unsigned parent;

	for (parent = PARENT_INDEX(i); i; i = parent, parent = PARENT_INDEX(i)) {
		if (pqueue->cmp(data, pqueue->list[parent]) <= 0)
			break;

		pqueue->list[i] = pqueue->list[parent];
		UPDATE_WATCHER_OFFSETS(pqueue, parent, i);
	}

	pqueue->list[i] = data;
	UPDATE_WATCHER_OFFSETS(pqueue, i, i);

	return i;
}

/*
 * Dequeue the most prior element
 */
PassiveInstRef *dequeue_priorQueue(PriorQueue *pqueue) {
	if (!pqueue->count)
		return NULL;

	PassiveInstRef *max = pqueue->list[0];

	unsigned count = --pqueue->count;
	PassiveInstRef *last = pqueue->list[count];

	unsigned i, child;
	for (i = 0, child = LEFT_CHILD_INDEX(i); child < count; i = child, child = LEFT_CHILD_INDEX(i)) {
		// set child to the greater of the two children
		if (child + 1 < count && pqueue->cmp(pqueue->list[child + 1], pqueue->list[child]) > 0)
			child++;

		// if last should be inserted above child
		if (pqueue->cmp(last, pqueue->list[child]) >= 0)
			break;

		pqueue->list[i] = pqueue->list[child];
		UPDATE_WATCHER_OFFSETS(pqueue, child, i);
	}

	pqueue->list[i] = last;
	UPDATE_WATCHER_OFFSETS(pqueue, i, i);

	return max;
}

/*
 * If the priority of an element has changed, one needs to place the element at a new position wrt. its new priority
 */
unsigned int rearrangeElement_priorQueue(PriorQueue *pqueue, unsigned int elemInd) {
	unsigned count = pqueue->count;
	ASSERT(elemInd < count);

	PassiveInstRef * data = pqueue->list[elemInd];

	unsigned i;
	for (i = PARENT_INDEX(elemInd); elemInd; elemInd = i, i = PARENT_INDEX(elemInd)) {
		if (pqueue->cmp(data, pqueue->list[i]) <= 0)
			break;

		pqueue->list[elemInd] = pqueue->list[i];
		UPDATE_WATCHER_OFFSETS(pqueue, i, elemInd);
	}
	for (i = LEFT_CHILD_INDEX(elemInd); i < count; elemInd = i, i = LEFT_CHILD_INDEX(elemInd)) {
		// set child to the greater of the two children
		if (i + 1 < count && pqueue->cmp(pqueue->list[i + 1], pqueue->list[i]) > 0)
			i++;

		if (pqueue->cmp(data, pqueue->list[i]) >= 0)
			break;

		pqueue->list[elemInd] = pqueue->list[i];
		UPDATE_WATCHER_OFFSETS(pqueue, i, elemInd);
	}

	pqueue->list[elemInd] = data;
	UPDATE_WATCHER_OFFSETS(pqueue, elemInd, elemInd);

	return elemInd;
}

/*
 * Deallocate a priority queue
 */
void delete_priorQueue(PriorQueue * pqueue) {
	iDQMemMan * mm = pqueue->mem;

	PassiveInstRef **pData;
	unsigned i;
	for (pData = pqueue->list, i = pqueue->count; i; pData++, i--)
		del(mm, *pData);

	DELETEN(mm, pqueue->list, pqueue->size);
	DELETE(mm, pqueue);
}
