/*
 This file is part of iDQ.

 iDQ is an instantiation-based DQBF (Dependency Quantified Boolean Formula) solver.

 Andreas Froehlich, Johannes Kepler University, Linz, Austria, 2012-2014.
 Gergely Kovasznai, Vienna University of Technology, Vienna, Austria, 2012-2014.

 iDQ is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or (at
 your option) any later version.

 iDQ is distributed in the hope that it will be useful, but
 WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with iDQ.  If not, see <http://www.gnu.org/licenses/>.

 This file incorporates work covered by the following copyright and
 permission notice:  

   DepQBF
   Copyright 2010-2014 Florian Lonsing, Johannes Kepler University, Linz, Austria.
   Copyright 2012 Aina Niemetz, Johannes Kepler University, Linz, Austria.
*/

#include <stdio.h>

#include "qbformula.h"
#include "idq.h"
#include "mem.h"
#include "assert.h"

/*
 * Initializes a QBFormula
 */
static void init_qbFormula(iDQMemMan * mm, QBFormula * qbf, int cntVars, int cntClauses) {
	qbf->cntVars = cntVars;
	qbf->cntClauses = cntClauses;
	qbf->cntDeletedClauses = 0;

	CNEWN(mm, qbf->pVars, cntVars + 1);
//	NEWN(mm, qbf->pClauses, cntClauses);
	CNEWN(mm, qbf->pClauses, cntClauses);
}

/*
 * Deletes a QBFormula
 */
void delete_qbFormula(iDQMemMan * mm, QBFormula * qbf) {
	if (!qbf->pClauses && !qbf->pVars)
		return;

	unsigned int i;

	Clause * pClause;
	for (pClause = qbf->pClauses, i = qbf->cntClauses; i; pClause++, i--)
		delete_clause(mm, pClause);
	DELETEN(mm, qbf->pClauses, qbf->cntClauses);

	Variable * pVar;
	for (pVar = qbf->pVars, i = qbf->cntVars + 1; i; pVar++, i--)
		delete_variable(mm, pVar);
	DELETEN(mm, qbf->pVars, qbf->cntVars + 1);
}

/*
 * Parsing input DQDIMACS file
 */

//static int lineNumber = 0;
static char ch;

// returns next character from file and updates current line number
static char nextChar(FILE * input) {
	return (ch = getc(input));
	/*	ch = getc(input);

	 if (ch == '\n')
	 lineNumber++;

	 return ch;*/
}

// checks whether a character is a whitespace
#define IS_WHITESPACE (ch == '\n' || ch == '\r' || ch == ' ' || ch == '\t')

// checks whether a character is a digit
#define IS_DIGIT (ch >= '0' && ch <= '9')

// reads the next sign, and returns it as 1 resp. -1
static int nextSign(FILE * input) {
	int s;

	if (ch == '-') {
		s = -1;
		nextChar(input);
		ASSERT(IS_DIGIT && ch != '0')
			ERRORMSG("expected positive digit after '-'\n");
	} else {
		s = 1;
		ASSERT(IS_DIGIT)
			ERRORMSG("expected '-' or digit");
	}

	return s;
}

// reads the next number
static int nextNumber(FILE * input) {
	int x = ch - '0';

	for (nextChar(input); IS_DIGIT; nextChar(input))
		x = 10 * x + (ch - '0');

	return x;
}

// copies the literas from the temporary memory to the current clause
static void pushClause(iDQ * idq, int clauseIndex) {
	Clause * currClause = idq->qbf->pClauses + clauseIndex;

	currClause->cntLits = COUNT_STACK(idq->tempLitStack);
	NEWN(idq->mm, currClause->lits, currClause->cntLits);

	LitID *pTempLit, *pEndTemp = idq->tempLitStack.top, *pLit;
	for (pTempLit = idq->tempLitStack.start, pLit = currClause->lits; pTempLit != pEndTemp; pTempLit++, pLit++)
		*pLit = *pTempLit;

	DELETE_STACK(idq->tempLitStack);
}

// copy the dependencies from the literal stack (which was abused for this) to the dependency list
static void pushDependencies(iDQ * idq, int varId) {
	Variable * currVar = idq->qbf->pVars + varId;

	currVar->cntDependencies = COUNT_STACK(idq->tempLitStack);
	NEWN(idq->mm, currVar->dependencies, currVar->cntDependencies);

	LitID *pTempLit, *pEndTemp = idq->tempLitStack.top;
	unsigned int *pDep;
	for (pTempLit = idq->tempLitStack.start, pDep = currVar->dependencies; pTempLit != pEndTemp; pTempLit++, pDep++) {
		ASSERT(*pTempLit > 0);
		*pDep = *pTempLit;
	}

	DELETE_STACK(idq->tempLitStack);
}

// main method for parsing a DQDIMACS file
void parse_dqdimacs(iDQ * idq) {
	FILE * file = idq->inputFile;
	QBFormula * qbf = idq->qbf;

	unsigned int currVar, currVar2, currClause, i;
	int currSign, currSign2;
	unsigned int cntVars, cntClauses, cntExistentials, cntUniversals;

// scanning for comments (c) or header (p)

	HEADER: nextChar(file);

// skip line whenever a comment is found
	if (ch == 'c') {
		while (nextChar(file) != '\n')
			ASSERT(ch != EOF)
				ERRORMSG("end of file in comment");
		goto HEADER;
	}

// if it's no comment it has to be the header ... otherwise return an error
	ASSERT(ch == 'p')
		ERRORMSG("expected header or comments");
	ASSERT(nextChar(file) == ' ')
		INVHEAD: ERRORMSG("expected 'p cnf <vars> <clauses>' header");
	while (nextChar(file) == ' ')
		;
	ASSERT(ch == 'c')
		goto INVHEAD;
	ASSERT(nextChar(file) == 'n')
		goto INVHEAD;
	ASSERT(nextChar(file) == 'f')
		goto INVHEAD;
	ASSERT(nextChar(file) == ' ')
		goto INVHEAD;

	while (nextChar(file) == ' ')
		;
	ASSERT(IS_DIGIT)
		goto INVHEAD;

// read the number of variables from the header line
	cntVars = nextNumber(file);

	ASSERT(ch == ' ')
		goto INVHEAD;
	while (nextChar(file) == ' ')
		;
	ASSERT(IS_DIGIT)
		goto INVHEAD;

// read the number of clauses from the header line
	cntClauses = nextNumber(file);

	ASSERT(IS_WHITESPACE)
		ERRORMSG("expected white space after header");

	init_qbFormula(idq->mm, qbf, cntVars, cntClauses);

	Variable * pVars = qbf->pVars;
	currVar = currClause = 0;

	cntExistentials = cntVars;
	cntUniversals = 0;

// read non-header stuff (quantifiers and clauses)
// variables not quantified explicitely will be existentially quantified implicitely in the outermost block (no dependencies)
	LIT: nextChar(file);
	if (IS_WHITESPACE)
		goto LIT;

// skip line whenever a comment is found
	if (ch == 'c') {
		while (nextChar(file) != '\n')
			ASSERT(ch != EOF)
				ERRORMSG("end of file in comment");
		goto LIT;
	}

// one or several universal quantified variables in this line
	if (ch == 'a') {
		if (idq->qdpll)
			qdpll_new_scope(idq->qdpll, QDPLL_QTYPE_FORALL);

		UNIVERSAL: while (nextChar(file) == ' ')
			;
		ASSERT(IS_DIGIT)
			ERRORMSG("expected digit (universal)");

		currVar = nextNumber(file);
		ASSERT(currVar <= cntVars)
			ERRORMSG("maximum variable index exceeded (universal)");
		ASSERT(IS_WHITESPACE)
			ERRORMSG("expected white space after literal (universal)");

		if (idq->qdpll)
			qdpll_add(idq->qdpll, currVar);

		if (currVar) {
			pVars[currVar].quantifier = QTYPE_FORALL;
			cntExistentials--;
			cntUniversals++;
			goto UNIVERSAL;
		} else
			goto LIT;
	}

// one or several existential quantified variables in this line
// will depend on all universal quantified variables that have already been defined
	if (ch == 'e') {
		if (idq->qdpll)
			qdpll_new_scope(idq->qdpll, QDPLL_QTYPE_EXISTS);

		EXISTENTIAL: while (nextChar(file) == ' ')
			;
		ASSERT(IS_DIGIT)
			ERRORMSG("expected digit (existential)");

		currVar = nextNumber(file);
		ASSERT(currVar <= cntVars)
			ERRORMSG("maximum variable index exceeded (existential)");
		ASSERT(IS_WHITESPACE)
			ERRORMSG("expected white space after literal (existential)");

		if (idq->qdpll)
			qdpll_add(idq->qdpll, currVar);

		if (currVar) {
			pVars[currVar].quantifier = QTYPE_EXISTS;
			for (i = 1; i <= cntVars; i++)
				if ((pVars + i)->quantifier == QTYPE_FORALL)
					PUSH_STACK(idq->tempLitStack, i);
			pushDependencies(idq, currVar);
			goto EXISTENTIAL;
		} else
			goto LIT;
	}

// a single existential quantified variable in this line
// will depend on the universal variables that are specified in the same line
	if (ch == 'd') {
		ASSERT(!idq->qdpll)
			ERRORMSG("dependency analysis can only be performed on QBF");

		while (nextChar(file) == ' ')
			;
		ASSERT(IS_DIGIT)
			ERRORMSG("expected digit (dependent existential)");

		currVar = nextNumber(file);
		ASSERT(currVar <= cntVars)
			ERRORMSG("maximum variable index exceeded (dependent existential)");
		pVars[currVar].quantifier = QTYPE_EXISTS;

		DEXISTENTIAL: while (nextChar(file) == ' ')
			;
		ASSERT(IS_DIGIT)
			ERRORMSG("expected digit (dependent existential)");

		i = nextNumber(file);
		ASSERT(i <= cntVars)
			ERRORMSG("maximum variable index exceeded (dependent existential)");
		ASSERT(IS_WHITESPACE)
			ERRORMSG("expected white space after literal (dependent existential)");

		if (i) {
			ASSERT((pVars + i)->quantifier == QTYPE_FORALL)
				ERRORMSG("dependency on existential variable not allowed!");

			PUSH_STACK(idq->tempLitStack, i);

			goto DEXISTENTIAL;
		} else {
			pushDependencies(idq, currVar);
			goto LIT;
		}
	}

// at the end of the file check whether last digit was 0
// and if the number of clauses read equals the number of clauses specified in the header
	if (ch == EOF) {
		ASSERT(!currVar)
			ERRORMSG("terminating zero missing");
		ASSERT(currClause >= cntClauses)
			ERRORMSG("too few (%d) clauses", currClause);
		ASSERT(currClause == cntClauses)
			ERRORMSG("too many (%d) clauses", currClause);

		qbf->cntVars = cntVars;
		qbf->cntClauses = cntClauses;
		qbf->cntExistentials = cntExistentials;
		qbf->cntUniversals = cntUniversals;

		return;
	}

// reads the sign of the next literal and checks if it is followed by a digit
	currSign = nextSign(file);

// read the digits, construct the variable they specify and check whether the variable is valid
	currVar = nextNumber(file);
	ASSERT(currVar <= cntVars)
		ERRORMSG("maximum variable index exceeded");
	ASSERT(IS_WHITESPACE || ch == '|')
		ERRORMSG("expected white space or '|' after literal");

// if the variable is not 0, add the literal to the clause ...
	if (currVar) {
		PUSH_STACK(idq->tempLitStack, currSign * currVar);

		if (idq->qdpll)
			qdpll_add(idq->qdpll, currVar);
	}
// ... otherwise reading the clause is finished and it can be added to the formula
	else {
		pushClause(idq, currClause++);
		if (idq->qdpll)
			qdpll_add(idq->qdpll, 0);
	}

	goto LIT;
}

/*
 * Dumps clause set in DQDIMACS format
 */
void dump_dqdimacs(iDQ * idq) {
	QBFormula * qbf = idq->qbf;
	FILE * out = idq->outputFile;

	unsigned int cnt, i, j;
	Variable * pVar;
	unsigned int * pVarId;
	Clause * pClause;
	LitID * pLit;

	fprintf(out, "c DQDIMACS dumped by iDQ\n");
	fprintf(out, "p cnf %d %d\n", qbf->cntVars, qbf->cntClauses);

	cnt = qbf->cntVars;

	if (qbf->cntUniversals) {
		fprintf(out, "a ");
		for (pVar = qbf->pVars + 1, i = 1; i <= cnt; pVar++, i++) {
			if (pVar->quantifier != QTYPE_FORALL)
				continue;
			fprintf(out, "%d ", i);
		}
		fprintf(out, "0\n");
	}

	if (qbf->cntExistentials) {
		for (pVar = qbf->pVars + 1, i = 1; i <= cnt; pVar++, i++) {
			if (pVar->quantifier != QTYPE_EXISTS)
				continue;
			fprintf(out, "d %d ", i);
			for (pVarId = pVar->dependencies, j = pVar->cntDependencies; j; pVarId++, j--)
				fprintf(out, "%d ", *pVarId);
			fprintf(out, "0\n");
		}
	}

	for (pClause = qbf->pClauses, i = qbf->cntClauses; i; pClause++, i--) {
		if (pClause->is_deleted)
			continue;
		for (pLit = pClause->lits, j = pClause->cntLits; j; pLit++, j--) {
			fprintf(out, "%d ", *pLit);
		}
		fprintf(out, "0\n");
	}
}

#define USE_MARTINAS_TPTP_ENCODING

#ifndef USE_MARTINAS_TPTP_ENCODING
#include "abformula.h"
#endif

/*
 * Dumps clause set in TPTP format
 */
void dump_tptp(iDQ * idq) {
	FILE * out = idq->outputFile;

#ifdef USE_MARTINAS_TPTP_ENCODING
	QBFormula * qbf = idq->qbf;

	unsigned int i, j, k;
	Variable * var;
	Clause * pClause;
	LitID * pLit;
	VarID varId;

	fprintf(out, "%% TPTP CNF dumped by iDQ\n");

	for (pClause = qbf->pClauses, i = 0; i < qbf->cntClauses; pClause++, i++) {
		if (!pClause->cntLits) {
			fprintf(out, "cnf(c_%d,plain,$false).\n", i);
			continue;
		}

		fprintf(out, "cnf(c_%d,plain,(", i);

		for (pLit = pClause->lits, j = 0; j < pClause->cntLits; pLit++, j++) {
			if (j)
				fprintf(out, " | ");

			if (*pLit < 0)
				fprintf(out, "~");

			varId = GET_VARID(*pLit);
			var = qbf->pVars + varId;
			if (var->quantifier == QTYPE_FORALL)
				fprintf(out, "p(X%d)", varId);
			else {
				fprintf(out, "e%d", varId);
				if (!var->cntDependencies)
					continue;

				fprintf(out, "(");
				for (k = 0; k < var->cntDependencies; k++) {
					if (k)
						fprintf(out, ",");
					fprintf(out, "X%d", var->dependencies[k]);
				}
				fprintf(out, ")");
			}
		}
		fprintf(out, ")).\n");
	}

	fprintf(out, "cnf(rule_true,axiom, p(1)).\n");
	fprintf(out, "cnf(rule_false,axiom, ~p(0)).\n");
#else
	AbFormula * abf;
	abf = idq->abf;

	unsigned int i, j, k;
	AbVariable * var;
	AbClause * pClause;
	LitID * pLit;
	VarID varId;
	LitID dep;
	IList * instMask;
	char is1stDep;

	fprintf(out, "%% TPTP CNF dumped by iDQ\n");

	for (pClause = abf->clauses, i = 0; i < abf->cntClauses; pClause++, i++) {
		if (!pClause->cntLits) continue;

		fprintf(out, "cnf(c_%d,plain,(", i);
		for (pLit = pClause->lits, j = 0; j < pClause->cntLits; pLit++, j++) {
			if (j)
			fprintf(out, " | ");

			if (*pLit < 0)
			fprintf(out, "~");

			varId = GET_VARID(*pLit);
			var = abf->vars + varId;

			fprintf(out, "e%d", varId);
			if (!countLits_iList(var->depVarMask))
			continue;

			instMask = TOP_STACK(pClause->instances)->instMask;

			is1stDep = 1;
			fprintf(out, "(");
			for (k = 1; k <= abf->cntUniversals; k++) {
				if (getLit_iList(var->depVarMask, k)) {
					if (!is1stDep)
					fprintf(out, ",");
					is1stDep = 0;

					if ((dep = getLit_iList(instMask, k)))
					fprintf(out, dep == 1 ? "1" : "0");
					else
					fprintf(out, "X%d", k);
				}
			}
			fprintf(out, ")");
		}
		fprintf(out, ")).\n");
	}
#endif
}
